# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing v0 related conversions from PyAnsys Geometry objects to gRPC messages."""

from typing import TYPE_CHECKING

import pint

from ansys.api.dbu.v0.admin_pb2 import BackendType as GRPCBackendType
from ansys.api.dbu.v0.dbumodels_pb2 import (
    DrivingDimension as GRPCDrivingDimension,
    EntityIdentifier,
    PartExportFormat as GRPCPartExportFormat,
)
from ansys.api.dbu.v0.drivingdimensions_pb2 import UpdateStatus as GRPCUpdateStatus
from ansys.api.geometry.v0.models_pb2 import (
    Arc as GRPCArc,
    Circle as GRPCCircle,
    CurveGeometry as GRPCCurveGeometry,
    Direction as GRPCDirection,
    Ellipse as GRPCEllipse,
    Frame as GRPCFrame,
    Geometries as GRPCGeometries,
    Line as GRPCLine,
    Material as GRPCMaterial,
    MaterialProperty as GRPCMaterialProperty,
    Plane as GRPCPlane,
    Point as GRPCPoint,
    Polygon as GRPCPolygon,
    Surface as GRPCSurface,
    SurfaceType as GRPCSurfaceType,
    Tessellation as GRPCTessellation,
    TessellationOptions as GRPCTessellationOptions,
    TrimmedCurve as GRPCTrimmedCurve,
    TrimmedSurface as GRPCTrimmedSurface,
)
from ansys.geometry.core.misc.checks import graphics_required

if TYPE_CHECKING:  # pragma: no cover
    import pyvista as pv

    from ansys.geometry.core.connection.backend import BackendType
    from ansys.geometry.core.designer.design import DesignFileFormat
    from ansys.geometry.core.designer.face import SurfaceType
    from ansys.geometry.core.materials.material import Material
    from ansys.geometry.core.materials.property import MaterialProperty
    from ansys.geometry.core.math.frame import Frame
    from ansys.geometry.core.math.plane import Plane
    from ansys.geometry.core.math.point import Point2D, Point3D
    from ansys.geometry.core.math.vector import UnitVector3D
    from ansys.geometry.core.misc.options import TessellationOptions
    from ansys.geometry.core.parameters.parameter import (
        Parameter,
        ParameterUpdateStatus,
    )
    from ansys.geometry.core.shapes.curves.curve import Curve
    from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve
    from ansys.geometry.core.shapes.surfaces.surface import Surface
    from ansys.geometry.core.shapes.surfaces.trimmed_surface import TrimmedSurface
    from ansys.geometry.core.sketch.arc import Arc
    from ansys.geometry.core.sketch.circle import SketchCircle
    from ansys.geometry.core.sketch.edge import SketchEdge
    from ansys.geometry.core.sketch.ellipse import SketchEllipse
    from ansys.geometry.core.sketch.face import SketchFace
    from ansys.geometry.core.sketch.polygon import Polygon
    from ansys.geometry.core.sketch.segment import SketchSegment


def from_point3d_to_grpc_point(point: "Point3D") -> GRPCPoint:
    """Convert a ``Point3D`` class to a point gRPC message.

    Parameters
    ----------
    point : Point3D
        Source point data.

    Returns
    -------
    GRPCPoint
        Geometry service gRPC point message. The unit is meters.
    """
    from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

    return GRPCPoint(
        x=point.x.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        y=point.y.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        z=point.z.m_as(DEFAULT_UNITS.SERVER_LENGTH),
    )


def from_grpc_point_to_point3d(point: GRPCPoint) -> "Point3D":
    """Convert a point gRPC message class to a ``Point3D`` class.

    Parameters
    ----------
    point : GRPCPoint
        Source point data.

    Returns
    -------
    Point3D
        Converted point.
    """
    from ansys.geometry.core.math.point import Point3D
    from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

    return Point3D(
        [point.x, point.y, point.z],
        DEFAULT_UNITS.SERVER_LENGTH,
    )


def from_point2d_to_grpc_point(plane: "Plane", point2d: "Point2D") -> GRPCPoint:
    """Convert a ``Point2D`` class to a point gRPC message.

    Parameters
    ----------
    plane : Plane
        Plane for positioning the 2D point.
    point : Point2D
        Source point data.

    Returns
    -------
    GRPCPoint
        Geometry service gRPC point message. The unit is meters.
    """
    from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

    point3d = plane.transform_point2d_local_to_global(point2d)
    return GRPCPoint(
        x=point3d.x.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        y=point3d.y.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        z=point3d.z.m_as(DEFAULT_UNITS.SERVER_LENGTH),
    )


def from_unit_vector_to_grpc_direction(unit_vector: "UnitVector3D") -> GRPCDirection:
    """Convert a ``UnitVector3D`` class to a unit vector gRPC message.

    Parameters
    ----------
    unit_vector : UnitVector3D
        Source vector data.

    Returns
    -------
    GRPCDirection
        Geometry service gRPC direction message.
    """
    return GRPCDirection(x=unit_vector.x, y=unit_vector.y, z=unit_vector.z)


def build_grpc_id(id: str) -> EntityIdentifier:
    """Build an EntityIdentifier gRPC message.

    Parameters
    ----------
    id : str
        Source ID.

    Returns
    -------
    EntityIdentifier
        Geometry service gRPC entity identifier message.
    """
    return EntityIdentifier(id=id)


def from_grpc_material_to_material(material: GRPCMaterial) -> "Material":
    """Convert a material gRPC message to a ``Material`` class.

    Parameters
    ----------
    material : GRPCMaterial
        Material gRPC message.

    Returns
    -------
    Material
        Converted material.
    """
    from ansys.geometry.core.materials.material import Material
    from ansys.geometry.core.materials.property import MaterialPropertyType
    from ansys.geometry.core.misc.units import UNITS

    properties = []
    density = pint.Quantity(0, UNITS.kg / UNITS.m**3)
    for property in material.material_properties:
        mp = from_grpc_material_property_to_material_property(property)
        properties.append(mp)
        if mp.type == MaterialPropertyType.DENSITY:
            density = mp.quantity

    return Material(material.name, density, properties)


def from_grpc_material_property_to_material_property(
    material_property: GRPCMaterialProperty,
) -> "MaterialProperty":
    """Convert a material property gRPC message to a ``MaterialProperty`` class.

    Parameters
    ----------
    material_property : GRPCMaterialProperty
        Material property gRPC message.

    Returns
    -------
    MaterialProperty
        Converted material property.
    """
    from ansys.geometry.core.materials.property import MaterialProperty, MaterialPropertyType

    try:
        mp_type = MaterialPropertyType.from_id(material_property.id)
    except ValueError:
        mp_type = material_property.id

    try:
        mp_quantity = pint.Quantity(material_property.value, material_property.units)
    except (
        pint.UndefinedUnitError,
        TypeError,
    ):
        mp_quantity = material_property.value

    return MaterialProperty(mp_type, material_property.display_name, mp_quantity)


def from_frame_to_grpc_frame(frame: "Frame") -> GRPCFrame:
    """Convert a ``Frame`` class to a frame gRPC message.

    Parameters
    ----------
    frame : Frame
        Source frame data.

    Returns
    -------
    GRPCFrame
        Geometry service gRPC frame message. The unit for the frame origin is meters.
    """
    return GRPCFrame(
        origin=from_point3d_to_grpc_point(frame.origin),
        dir_x=from_unit_vector_to_grpc_direction(frame.direction_x),
        dir_y=from_unit_vector_to_grpc_direction(frame.direction_y),
    )


def from_grpc_frame_to_frame(frame: GRPCFrame) -> "Frame":
    """Convert a frame gRPC message to a ``Frame`` class.

    Parameters
    ----------
    frame : GRPCFrame
        Source frame data.

    Returns
    -------
    Frame
        Converted frame.
    """
    from ansys.geometry.core.math.frame import Frame
    from ansys.geometry.core.math.point import Point3D
    from ansys.geometry.core.math.vector import UnitVector3D
    from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

    return Frame(
        Point3D(
            input=[
                frame.origin.x,
                frame.origin.y,
                frame.origin.z,
            ],
            unit=DEFAULT_UNITS.SERVER_LENGTH,
        ),
        UnitVector3D(
            input=[
                frame.dir_x.x,
                frame.dir_x.y,
                frame.dir_x.z,
            ]
        ),
        UnitVector3D(
            input=[
                frame.dir_y.x,
                frame.dir_y.y,
                frame.dir_y.z,
            ]
        ),
    )


def from_plane_to_grpc_plane(plane: "Plane") -> GRPCPlane:
    """Convert a ``Plane`` class to a plane gRPC message.

    Parameters
    ----------
    plane : Plane
        Source plane data.

    Returns
    -------
    GRPCPlane
        Geometry service gRPC plane message. The unit is meters.
    """
    return GRPCPlane(
        frame=GRPCFrame(
            origin=from_point3d_to_grpc_point(plane.origin),
            dir_x=from_unit_vector_to_grpc_direction(plane.direction_x),
            dir_y=from_unit_vector_to_grpc_direction(plane.direction_y),
        )
    )


@graphics_required
def from_grpc_tess_to_pd(tess: GRPCTessellation) -> "pv.PolyData":
    """Convert a ``Tessellation`` to ``pyvista.PolyData``."""
    # lazy imports here to improve initial load
    import numpy as np
    import pyvista as pv

    return pv.PolyData(var_inp=np.array(tess.vertices).reshape(-1, 3), faces=tess.faces)


def from_tess_options_to_grpc_tess_options(
    options: "TessellationOptions",
) -> GRPCTessellationOptions:
    """Convert a ``TessellationOptions`` class to a tessellation options gRPC message.

    Parameters
    ----------
    options : TessellationOptions
        Source tessellation options data.

    Returns
    -------
    GRPCTessellationOptions
        Geometry service gRPC tessellation options message.
    """
    return GRPCTessellationOptions(
        surface_deviation=options.surface_deviation,
        angle_deviation=options.angle_deviation,
        maximum_aspect_ratio=options.max_aspect_ratio,
        maximum_edge_length=options.max_edge_length,
        watertight=options.watertight,
    )


def from_sketch_shapes_to_grpc_geometries(
    plane: "Plane",
    edges: list["SketchEdge"],
    faces: list["SketchFace"],
    only_one_curve: bool = False,
) -> GRPCGeometries:
    """Convert lists of ``SketchEdge`` and ``SketchFace`` to a gRPC message.

    Parameters
    ----------
    plane : Plane
        Plane for positioning the 2D sketches.
    edges : list[SketchEdge]
        Source edge data.
    faces : list[SketchFace]
        Source face data.
    only_one_curve : bool, default: False
        Whether to project one curve of the whole set of geometries to
        enhance performance.

    Returns
    -------
    GRPCGeometries
        Geometry service gRPC geometries message. The unit is meters.
    """
    from ansys.geometry.core.sketch.circle import SketchCircle
    from ansys.geometry.core.sketch.ellipse import SketchEllipse
    from ansys.geometry.core.sketch.polygon import Polygon

    geometries = GRPCGeometries()

    converted_sketch_edges = from_sketch_edges_to_grpc_geometries(edges, plane)
    geometries.lines.extend(converted_sketch_edges[0])
    geometries.arcs.extend(converted_sketch_edges[1])

    for face in faces:
        if isinstance(face, SketchCircle):
            geometries.circles.append(from_sketch_circle_to_grpc_circle(face, plane))
        if isinstance(face, SketchEllipse):
            geometries.ellipses.append(from_sketch_ellipse_to_grpc_ellipse(face, plane))
        if isinstance(face, Polygon):
            geometries.polygons.append(from_sketch_polygon_to_grpc_polygon(face, plane))
        else:
            converted_face_edges = from_sketch_edges_to_grpc_geometries(face.edges, plane)
            geometries.lines.extend(converted_face_edges[0])
            geometries.arcs.extend(converted_face_edges[1])

    if only_one_curve:
        one_curve_geometry = GRPCGeometries()
        if len(geometries.lines) > 0:
            one_curve_geometry.lines.append(geometries.lines[0])
        elif len(geometries.arcs) > 0:
            one_curve_geometry.arcs.append(geometries.arcs[0])
        elif len(geometries.circles) > 0:
            one_curve_geometry.circles.append(geometries.circles[0])
        elif len(geometries.ellipses) > 0:
            one_curve_geometry.ellipses.append(geometries.ellipses[0])
        elif len(geometries.polygons) > 0:
            one_curve_geometry.polygons.append(geometries.polygons[0])
        return one_curve_geometry

    else:
        return geometries


def from_sketch_edges_to_grpc_geometries(
    edges: list["SketchEdge"],
    plane: "Plane",
) -> tuple[list[GRPCLine], list[GRPCArc]]:
    """Convert a list of ``SketchEdge`` to a gRPC message.

    Parameters
    ----------
    edges : list[SketchEdge]
        Source edge data.
    plane : Plane
        Plane for positioning the 2D sketches.

    Returns
    -------
    tuple[list[GRPCLine], list[GRPCArc]]
        Geometry service gRPC line and arc messages. The unit is meters.
    """
    from ansys.geometry.core.sketch.arc import Arc
    from ansys.geometry.core.sketch.segment import SketchSegment

    arcs = []
    segments = []
    for edge in edges:
        if isinstance(edge, SketchSegment):
            segments.append(from_sketch_segment_to_grpc_line(edge, plane))
        elif isinstance(edge, Arc):
            arcs.append(from_sketch_arc_to_grpc_arc(edge, plane))

    return (segments, arcs)


def from_sketch_arc_to_grpc_arc(arc: "Arc", plane: "Plane") -> GRPCArc:
    """Convert an ``Arc`` class to an arc gRPC message.

    Parameters
    ----------
    arc : Arc
        Source arc data.
    plane : Plane
        Plane for positioning the arc within.

    Returns
    -------
    GRPCArc
        Geometry service gRPC arc message. The unit is meters.
    """
    axis = (
        from_unit_vector_to_grpc_direction(plane.direction_z)
        if not arc.is_clockwise
        else from_unit_vector_to_grpc_direction(-plane.direction_z)
    )

    return GRPCArc(
        center=from_point2d_to_grpc_point(plane, arc.center),
        start=from_point2d_to_grpc_point(plane, arc.start),
        end=from_point2d_to_grpc_point(plane, arc.end),
        axis=axis,
    )


def from_sketch_ellipse_to_grpc_ellipse(ellipse: "SketchEllipse", plane: "Plane") -> GRPCEllipse:
    """Convert a ``SketchEllipse`` class to an ellipse gRPC message.

    Parameters
    ----------
    ellipse : SketchEllipse
        Source ellipse data.

    Returns
    -------
    GRPCEllipse
        Geometry service gRPC ellipse message. The unit is meters.
    """
    from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

    return GRPCEllipse(
        center=from_point2d_to_grpc_point(plane, ellipse.center),
        majorradius=ellipse.major_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        minorradius=ellipse.minor_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        angle=ellipse.angle.m_as(DEFAULT_UNITS.SERVER_ANGLE),
    )


def from_sketch_circle_to_grpc_circle(circle: "SketchCircle", plane: "Plane") -> GRPCCircle:
    """Convert a ``SketchCircle`` class to a circle gRPC message.

    Parameters
    ----------
    circle : SketchCircle
        Source circle data.
    plane : Plane
        Plane for positioning the circle.

    Returns
    -------
    GRPCCircle
        Geometry service gRPC circle message. The unit is meters.
    """
    from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

    return GRPCCircle(
        center=from_point2d_to_grpc_point(plane, circle.center),
        radius=circle.radius.m_as(DEFAULT_UNITS.SERVER_LENGTH),
    )


def from_sketch_polygon_to_grpc_polygon(polygon: "Polygon", plane: "Plane") -> GRPCPolygon:
    """Convert a ``Polygon`` class to a polygon gRPC message.

    Parameters
    ----------
    polygon : Polygon
        Source polygon data.

    Returns
    -------
    GRPCPolygon
        Geometry service gRPC polygon message. The unit is meters.
    """
    from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

    return GRPCPolygon(
        center=from_point2d_to_grpc_point(plane, polygon.center),
        radius=polygon.inner_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH),
        numberofsides=polygon.n_sides,
        angle=polygon.angle.m_as(DEFAULT_UNITS.SERVER_ANGLE),
    )


def from_sketch_segment_to_grpc_line(segment: "SketchSegment", plane: "Plane") -> GRPCLine:
    """Convert a ``Segment`` class to a line gRPC message.

    Parameters
    ----------
    segment : SketchSegment
        Source segment data.

    Returns
    -------
    GRPCLine
        Geometry service gRPC line message. The unit is meters.
    """
    return GRPCLine(
        start=from_point2d_to_grpc_point(plane, segment.start),
        end=from_point2d_to_grpc_point(plane, segment.end),
    )


def from_trimmed_curve_to_grpc_trimmed_curve(curve: "TrimmedCurve") -> GRPCTrimmedCurve:
    """Convert a ``TrimmedCurve`` to a trimmed curve gRPC message.

    Parameters
    ----------
    curve : TrimmedCurve
        Curve to convert.

    Returns
    -------
    GRPCTrimmedCurve
        Geometry service gRPC ``TrimmedCurve`` message.
    """
    curve_geometry = from_curve_to_grpc_curve(curve.geometry)
    i_start = curve.interval.start
    i_end = curve.interval.end

    return GRPCTrimmedCurve(
        curve=curve_geometry,
        interval_start=i_start,
        interval_end=i_end,
    )


def from_curve_to_grpc_curve(curve: "Curve") -> GRPCCurveGeometry:
    """Convert a ``Curve`` object to a curve gRPC message.

    Parameters
    ----------
    curve : Curve
        Curve to convert.

    Returns
    -------
    GRPCCurve
        Return ``Curve`` as a ``ansys.api.geometry.CurveGeometry`` message.
    """
    from ansys.geometry.core.shapes.curves.circle import Circle
    from ansys.geometry.core.shapes.curves.ellipse import Ellipse
    from ansys.geometry.core.shapes.curves.line import Line

    grpc_curve = None

    if isinstance(curve, Line):
        origin = from_point3d_to_grpc_point(curve.origin)
        direction = from_unit_vector_to_grpc_direction(curve.direction)
        grpc_curve = GRPCCurveGeometry(origin=origin, direction=direction)
    elif isinstance(curve, (Circle, Ellipse)):
        origin = from_point3d_to_grpc_point(curve.origin)
        reference = from_unit_vector_to_grpc_direction(curve.dir_x)
        axis = from_unit_vector_to_grpc_direction(curve.dir_z)

        if isinstance(curve, Circle):
            grpc_curve = GRPCCurveGeometry(
                origin=origin, reference=reference, axis=axis, radius=curve.radius.m
            )
        elif isinstance(curve, Ellipse):
            grpc_curve = GRPCCurveGeometry(
                origin=origin,
                reference=reference,
                axis=axis,
                major_radius=curve.major_radius.m,
                minor_radius=curve.minor_radius.m,
            )
    else:
        raise ValueError(f"Unsupported curve type: {type(curve)}")

    return grpc_curve


def from_grpc_curve_to_curve(curve: GRPCCurveGeometry) -> "Curve":
    """Convert a curve gRPC message to a ``Curve``.

    Parameters
    ----------
    curve : GRPCCurve
        Geometry service gRPC curve message.

    Returns
    -------
    Curve
        Resulting converted curve.
    """
    from ansys.geometry.core.math.point import Point3D
    from ansys.geometry.core.math.vector import UnitVector3D
    from ansys.geometry.core.shapes.curves.circle import Circle
    from ansys.geometry.core.shapes.curves.ellipse import Ellipse
    from ansys.geometry.core.shapes.curves.line import Line

    origin = Point3D([curve.origin.x, curve.origin.y, curve.origin.z])
    try:
        reference = UnitVector3D([curve.reference.x, curve.reference.y, curve.reference.z])
        axis = UnitVector3D([curve.axis.x, curve.axis.y, curve.axis.z])
    except ValueError:
        # curve will be a line
        pass
    if curve.radius != 0:
        result = Circle(origin, curve.radius, reference, axis)
    elif curve.major_radius != 0 and curve.minor_radius != 0:
        result = Ellipse(origin, curve.major_radius, curve.minor_radius, reference, axis)
    elif curve.direction is not None:
        result = Line(
            origin,
            UnitVector3D(
                [
                    curve.direction.x,
                    curve.direction.y,
                    curve.direction.z,
                ]
            ),
        )
    else:
        result = None

    return result


def from_trimmed_surface_to_grpc_trimmed_surface(
    trimmed_surface: "TrimmedSurface",
) -> GRPCTrimmedSurface:
    """Convert a ``TrimmedSurface`` to a trimmed surface gRPC message.

    Parameters
    ----------
    trimmed_surface : TrimmedSurface
        Surface to convert.

    Returns
    -------
    GRPCTrimmedSurface
        Geometry service gRPC ``TrimmedSurface`` message.
    """
    surface_geometry, surface_type = from_surface_to_grpc_surface(trimmed_surface.geometry)

    return GRPCTrimmedSurface(
        surface=surface_geometry,
        type=surface_type,
        u_min=trimmed_surface.box_uv.interval_u.start,
        u_max=trimmed_surface.box_uv.interval_u.end,
        v_min=trimmed_surface.box_uv.interval_v.start,
        v_max=trimmed_surface.box_uv.interval_v.end,
    )


def from_surface_to_grpc_surface(surface: "Surface") -> tuple[GRPCSurface, GRPCSurfaceType]:
    """Convert a ``Surface`` object to a surface gRPC message.

    Parameters
    ----------
    surface : Surface
        Surface to convert.

    Returns
    -------
    GRPCSurface
        Return ``Surface`` as a ``ansys.api.geometry.Surface`` message.
    GRPCSurfaceType
        Return the grpc surface type of ``Surface``.
    """
    from ansys.geometry.core.shapes.surfaces.cone import Cone
    from ansys.geometry.core.shapes.surfaces.cylinder import Cylinder
    from ansys.geometry.core.shapes.surfaces.plane import PlaneSurface
    from ansys.geometry.core.shapes.surfaces.sphere import Sphere
    from ansys.geometry.core.shapes.surfaces.torus import Torus

    grpc_surface = None
    surface_type = None
    origin = from_point3d_to_grpc_point(surface.origin)
    reference = from_unit_vector_to_grpc_direction(surface.dir_x)
    axis = from_unit_vector_to_grpc_direction(surface.dir_z)

    if isinstance(surface, PlaneSurface):
        grpc_surface = GRPCSurface(origin=origin, reference=reference, axis=axis)
        surface_type = GRPCSurfaceType.SURFACETYPE_PLANE
    elif isinstance(surface, Sphere):
        grpc_surface = GRPCSurface(
            origin=origin, reference=reference, axis=axis, radius=surface.radius.m
        )
        surface_type = GRPCSurfaceType.SURFACETYPE_SPHERE
    elif isinstance(surface, Cylinder):
        grpc_surface = GRPCSurface(
            origin=origin, reference=reference, axis=axis, radius=surface.radius.m
        )
        surface_type = GRPCSurfaceType.SURFACETYPE_CYLINDER
    elif isinstance(surface, Cone):
        grpc_surface = GRPCSurface(
            origin=origin,
            reference=reference,
            axis=axis,
            radius=surface.radius.m,
            half_angle=surface.half_angle.m,
        )
        surface_type = GRPCSurfaceType.SURFACETYPE_CONE
    elif isinstance(surface, Torus):
        grpc_surface = GRPCSurface(
            origin=origin,
            reference=reference,
            axis=axis,
            major_radius=surface.major_radius.m,
            minor_radius=surface.minor_radius.m,
        )
        surface_type = GRPCSurfaceType.SURFACETYPE_TORUS

    return grpc_surface, surface_type


def from_grpc_surface_to_surface(surface: GRPCSurface, surface_type: "SurfaceType") -> "Surface":
    """Convert a surface gRPC message to a ``Surface`` class.

    Parameters
    ----------
    surface : GRPCSurface
        Geometry service gRPC surface message.

    Returns
    -------
    Surface
        Resulting converted surface.
    """
    from ansys.geometry.core.designer.face import SurfaceType
    from ansys.geometry.core.math.vector import UnitVector3D
    from ansys.geometry.core.shapes.surfaces.cone import Cone
    from ansys.geometry.core.shapes.surfaces.cylinder import Cylinder
    from ansys.geometry.core.shapes.surfaces.plane import PlaneSurface
    from ansys.geometry.core.shapes.surfaces.sphere import Sphere
    from ansys.geometry.core.shapes.surfaces.torus import Torus

    origin = from_grpc_point_to_point3d(surface.origin)
    axis = UnitVector3D([surface.axis.x, surface.axis.y, surface.axis.z])
    reference = UnitVector3D([surface.reference.x, surface.reference.y, surface.reference.z])

    if surface_type == SurfaceType.SURFACETYPE_CONE:
        result = Cone(origin, surface.radius, surface.half_angle, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_CYLINDER:
        result = Cylinder(origin, surface.radius, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_SPHERE:
        result = Sphere(origin, surface.radius, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_TORUS:
        result = Torus(origin, surface.major_radius, surface.minor_radius, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_PLANE:
        result = PlaneSurface(origin, reference, axis)
    else:
        result = None
    return result


def from_grpc_backend_type_to_backend_type(
    grpc_backend_type: GRPCBackendType,
) -> "BackendType":
    """Convert a gRPC backend type to a backend type.

    Parameters
    ----------
    backend_type : GRPCBackendType
        Source backend type.

    Returns
    -------
    BackendType
        Converted backend type.
    """
    from ansys.geometry.core.connection.backend import BackendType

    # Map the gRPC backend type to the corresponding BackendType
    backend_type = None

    if grpc_backend_type == GRPCBackendType.DISCOVERY:
        backend_type = BackendType.DISCOVERY
    elif grpc_backend_type == GRPCBackendType.SPACECLAIM:
        backend_type = BackendType.SPACECLAIM
    elif grpc_backend_type == GRPCBackendType.WINDOWS_DMS:
        backend_type = BackendType.WINDOWS_SERVICE
    elif grpc_backend_type == GRPCBackendType.LINUX_DMS:
        backend_type = BackendType.LINUX_SERVICE
    elif grpc_backend_type == GRPCBackendType.CORE_SERVICE_LINUX:
        backend_type = BackendType.CORE_LINUX
    elif grpc_backend_type == GRPCBackendType.CORE_SERVICE_WINDOWS:
        backend_type = BackendType.CORE_WINDOWS
    elif grpc_backend_type == GRPCBackendType.DISCOVERY_HEADLESS:
        backend_type = BackendType.DISCOVERY_HEADLESS
    else:
        raise ValueError(f"Invalid backend type: {grpc_backend_type}")

    return backend_type


def from_grpc_driving_dimension_to_driving_dimension(
    driving_dimension: GRPCDrivingDimension,
) -> "Parameter":
    """Convert a gRPC driving dimension to a driving dimension object.

    Parameters
    ----------
    driving_dimension : GRPCDrivingDimension
        Source driving dimension type.

    Returns
    -------
    Parameter
        Converted driving dimension.
    """
    from ansys.geometry.core.parameters.parameter import Parameter, ParameterType

    return Parameter(
        id=driving_dimension.id,
        name=driving_dimension.name,
        dimension_type=ParameterType(driving_dimension.dimension_type),
        dimension_value=driving_dimension.dimension_value,
    )


def from_driving_dimension_to_grpc_driving_dimension(
    driving_dimension: "Parameter",
) -> GRPCDrivingDimension:
    """Convert a driving dimension object to a gRPC driving dimension.

    Parameters
    ----------
    driving_dimension : Parameter
        Source driving dimension type.

    Returns
    -------
    GRPCDrivingDimension
        Converted driving dimension.
    """
    return GRPCDrivingDimension(
        id=driving_dimension.id,
        name=driving_dimension.name,
        dimension_type=driving_dimension.dimension_type.value,
        dimension_value=driving_dimension.dimension_value,
    )


def from_grpc_update_status_to_parameter_update_status(
    update_status: GRPCUpdateStatus,
) -> "ParameterUpdateStatus":
    """Convert a gRPC update status to a parameter update status.

    Parameters
    ----------
    update_status : GRPCUpdateStatus
        Source update status.

    Returns
    -------
    ParameterUpdateStatus
        Converted update status.
    """
    from ansys.geometry.core.parameters.parameter import ParameterUpdateStatus

    status_mapping = {
        GRPCUpdateStatus.SUCCESS: ParameterUpdateStatus.SUCCESS,
        GRPCUpdateStatus.FAILURE: ParameterUpdateStatus.FAILURE,
        GRPCUpdateStatus.CONSTRAINED_PARAMETERS: ParameterUpdateStatus.CONSTRAINED_PARAMETERS,
    }
    return status_mapping.get(update_status, ParameterUpdateStatus.UNKNOWN)


def from_design_file_format_to_grpc_part_export_format(
    design_file_format: "DesignFileFormat",
) -> GRPCPartExportFormat:
    """Convert from a DesignFileFormat object to a gRPC PartExportFormat one.

    Parameters
    ----------
    design_file_format : DesignFileFormat
        The file format desired

    Returns
    -------
    GRPCPartExportFormat
        Converted gRPC Part format
    """
    from ansys.geometry.core.designer.design import DesignFileFormat

    if design_file_format == DesignFileFormat.SCDOCX:
        return GRPCPartExportFormat.PARTEXPORTFORMAT_SCDOCX
    elif design_file_format == DesignFileFormat.PARASOLID_TEXT:
        return GRPCPartExportFormat.PARTEXPORTFORMAT_PARASOLID_TEXT
    elif design_file_format == DesignFileFormat.PARASOLID_BIN:
        return GRPCPartExportFormat.PARTEXPORTFORMAT_PARASOLID_BINARY
    elif design_file_format == DesignFileFormat.FMD:
        return GRPCPartExportFormat.PARTEXPORTFORMAT_FMD
    elif design_file_format == DesignFileFormat.STEP:
        return GRPCPartExportFormat.PARTEXPORTFORMAT_STEP
    elif design_file_format == DesignFileFormat.IGES:
        return GRPCPartExportFormat.PARTEXPORTFORMAT_IGES
    elif design_file_format == DesignFileFormat.PMDB:
        return GRPCPartExportFormat.PARTEXPORTFORMAT_PMDB
    elif design_file_format == DesignFileFormat.STRIDE:
        return GRPCPartExportFormat.PARTEXPORTFORMAT_STRIDE
    elif design_file_format == DesignFileFormat.DISCO:
        return GRPCPartExportFormat.PARTEXPORTFORMAT_DISCO
    else:
        return None


def from_material_to_grpc_material(
    material: "Material",
) -> GRPCMaterial:
    """Convert a ``Material`` class to a material gRPC message.

    Parameters
    ----------
    material : Material
        Source material data.

    Returns
    -------
    GRPCMaterial
        Geometry service gRPC material message.
    """
    return GRPCMaterial(
        name=material.name,
        material_properties=[
            GRPCMaterialProperty(
                id=property.type.value,
                display_name=property.name,
                value=property.quantity.m,
                units=format(property.quantity.units),
            )
            for property in material.properties.values()
        ],
    )



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the designs service implementation (abstraction layer)."""

from abc import ABC, abstractmethod

import grpc


class GRPCDesignsService(ABC):  # pragma: no cover
    """Designs service for gRPC communication with the Geometry server.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    def __init__(self, channel: grpc.Channel):
        """Initialize the GRPCDesignsService class."""
        pass

    @abstractmethod
    def open(self, **kwargs) -> dict:
        """Open a design on the service."""
        pass

    @abstractmethod
    def new(self, **kwargs) -> dict:
        """Create a new design."""
        pass

    @abstractmethod
    def close(self, **kwargs) -> dict:
        """Close the currently open design."""
        pass

    @abstractmethod
    def put_active(self, **kwargs) -> dict:
        """Activate an already opened design on the service."""
        pass

    @abstractmethod
    def save_as(self, **kwargs) -> dict:
        """Create a new design."""
        pass

    @abstractmethod
    def download_export(self, **kwargs) -> dict:
        """Download and export a design into a certain format."""
        pass

    @abstractmethod
    def stream_download_export(self, **kwargs) -> dict:
        """Download and export a design into a certain format."""
        pass

    @abstractmethod
    def insert(self, **kwargs) -> dict:
        """Insert a part/component/design into an existing design."""
        pass

    @abstractmethod
    def get_active(self, **kwargs) -> dict:
        """Get the active design on the service."""
        pass



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides tools for repairing bodies."""

from typing import TYPE_CHECKING

from ansys.geometry.core.connection import GrpcClient
from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.misc.auxiliary import (
    get_bodies_from_ids,
    get_design_from_body,
    get_edges_from_ids,
    get_faces_from_ids,
)
from ansys.geometry.core.misc.checks import (
    check_type,
    check_type_all_elements_in_iterable,
    min_backend_version,
)
from ansys.geometry.core.tools.check_geometry import GeometryIssue, InspectResult
from ansys.geometry.core.tools.problem_areas import (
    DuplicateFaceProblemAreas,
    ExtraEdgeProblemAreas,
    InexactEdgeProblemAreas,
    InterferenceProblemAreas,
    MissingFaceProblemAreas,
    ShortEdgeProblemAreas,
    SmallFaceProblemAreas,
    SplitEdgeProblemAreas,
    StitchFaceProblemAreas,
    UnsimplifiedFaceProblemAreas,
)
from ansys.geometry.core.tools.repair_tool_message import RepairToolMessage
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.modeler import Modeler


class RepairTools:
    """Repair tools for PyAnsys Geometry."""

    def __init__(self, grpc_client: GrpcClient, modeler: "Modeler"):
        """Initialize a new instance of the ``RepairTools`` class."""
        self._modeler = modeler
        self._grpc_client = grpc_client

    @protect_grpc
    def find_split_edges(
        self, bodies: list["Body"], angle: Real = 0.0, length: Real = 0.0
    ) -> list[SplitEdgeProblemAreas]:
        """Find split edges in the given list of bodies.

        This method finds the split edge problem areas and returns a list of split edge
        problem areas objects.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that split edges are investigated on.
        angle : Real
            The maximum angle between edges.
        length : Real
            The maximum length of the edges.

        Returns
        -------
        list[SplitEdgeProblemAreas]
            List of objects representing split edge problem areas.
        """
        if not bodies:
            return []

        body_ids = [body.id for body in bodies]

        problem_areas_response = self._grpc_client.services.repair_tools.find_split_edges(
            bodies_or_faces=body_ids, angle=angle, distance=length
        )

        parent_design = get_design_from_body(bodies[0])
        return [
            SplitEdgeProblemAreas(
                f"{res['id']}",
                self._grpc_client,
                get_edges_from_ids(parent_design, res["edges"]),
            )
            for res in problem_areas_response["problems"]
        ]

    @protect_grpc
    def find_extra_edges(self, bodies: list["Body"]) -> list[ExtraEdgeProblemAreas]:
        """Find the extra edges in the given list of bodies.

        This method find the extra edge problem areas and returns a list of extra edge
        problem areas objects.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that extra edges are investigated on.

        Returns
        -------
        list[ExtraEdgeProblemArea]
            List of objects representing extra edge problem areas.
        """
        if not bodies:
            return []

        body_ids = [body.id for body in bodies]
        problem_areas_response = self._grpc_client.services.repair_tools.find_extra_edges(
            selection=body_ids
        )
        parent_design = get_design_from_body(bodies[0])

        return [
            ExtraEdgeProblemAreas(
                f"{res['id']}",
                self._grpc_client,
                get_edges_from_ids(parent_design, res["edges"]),
            )
            for res in problem_areas_response["problems"]
        ]

    @protect_grpc
    def find_inexact_edges(self, bodies: list["Body"]) -> list[InexactEdgeProblemAreas]:
        """Find inexact edges in the given list of bodies.

        This method find the inexact edge problem areas and returns a list of inexact
        edge problem areas objects.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that inexact edges are investigated on.

        Returns
        -------
        list[InExactEdgeProblemArea]
            List of objects representing inexact edge problem areas.
        """
        if not bodies:
            return []

        body_ids = [body.id for body in bodies]
        problem_areas_response = self._grpc_client.services.repair_tools.find_inexact_edges(
            selection=body_ids
        )

        parent_design = get_design_from_body(bodies[0])

        return [
            InexactEdgeProblemAreas(
                f"{res['id']}",
                self._grpc_client,
                get_edges_from_ids(parent_design, res["edges"]),
            )
            for res in problem_areas_response["problems"]
        ]

    @protect_grpc
    def find_short_edges(
        self, bodies: list["Body"], length: Real = 0.0
    ) -> list[ShortEdgeProblemAreas]:
        """Find the short edge problem areas.

        This method finds the short edge problem areas and returns a list of
        these objects.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that short edges are investigated on.

        Returns
        -------
        list[ShortEdgeProblemAreas]
            List of objects representing short edge problem areas.
        """
        if not bodies:
            return []

        body_ids = [body.id for body in bodies]

        problem_areas_response = self._grpc_client.services.repair_tools.find_short_edges(
            selection=body_ids, length=length
        )

        parent_design = get_design_from_body(bodies[0])
        return [
            ShortEdgeProblemAreas(
                f"{res['id']}",
                self._grpc_client,
                get_edges_from_ids(parent_design, res["edges"]),
            )
            for res in problem_areas_response["problems"]
        ]

    @protect_grpc
    def find_duplicate_faces(self, bodies: list["Body"]) -> list[DuplicateFaceProblemAreas]:
        """Find the duplicate face problem areas.

        This method finds the duplicate face problem areas and returns a list of
        duplicate face problem areas objects.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that duplicate faces are investigated on.

        Returns
        -------
        list[DuplicateFaceProblemAreas]
            List of objects representing duplicate face problem areas.
        """
        if not bodies:
            return []

        body_ids = [body.id for body in bodies]
        problem_areas_response = self._grpc_client.services.repair_tools.find_duplicate_faces(
            faces=body_ids
        )

        parent_design = get_design_from_body(bodies[0])
        return [
            DuplicateFaceProblemAreas(
                f"{res['id']}",
                self._grpc_client,
                get_faces_from_ids(parent_design, res["faces"]),
            )
            for res in problem_areas_response["problems"]
        ]

    @protect_grpc
    def find_missing_faces(self, bodies: list["Body"]) -> list[MissingFaceProblemAreas]:
        """Find the missing faces.

        This method find the missing face problem areas and returns a list of missing
        face problem areas objects.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that missing faces are investigated on.

        Returns
        -------
        list[MissingFaceProblemAreas]
            List of objects representing missing face problem areas.
        """
        if not bodies:
            return []
        body_ids = [body.id for body in bodies]
        problem_areas_response = self._grpc_client.services.repair_tools.find_missing_faces(
            faces=body_ids
        )
        parent_design = get_design_from_body(bodies[0])

        return [
            MissingFaceProblemAreas(
                f"{res['id']}",
                self._grpc_client,
                get_edges_from_ids(parent_design, res["edges"]),
            )
            for res in problem_areas_response["problems"]
        ]

    @protect_grpc
    def find_small_faces(self, bodies: list["Body"]) -> list[SmallFaceProblemAreas]:
        """Find the small face problem areas.

        This method finds and returns a list of ids of small face problem areas
        objects.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that small faces are investigated on.

        Returns
        -------
        list[SmallFaceProblemAreas]
            List of objects representing small face problem areas.
        """
        if not bodies:
            return []

        body_ids = [body.id for body in bodies]
        problem_areas_response = self._grpc_client.services.repair_tools.find_small_faces(
            selection=body_ids
        )
        parent_design = get_design_from_body(bodies[0])

        return [
            SmallFaceProblemAreas(
                f"{res['id']}",
                self._grpc_client,
                get_faces_from_ids(parent_design, res["faces"]),
            )
            for res in problem_areas_response["problems"]
        ]

    @protect_grpc
    def find_stitch_faces(self, bodies: list["Body"]) -> list[StitchFaceProblemAreas]:
        """Return the list of stitch face problem areas.

        This method find the stitch face problem areas and returns a list of ids of stitch face
        problem areas objects.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that stitchable faces are investigated on.

        Returns
        -------
        list[StitchFaceProblemAreas]
            List of objects representing stitch face problem areas.
        """
        body_ids = [body.id for body in bodies]
        problem_areas_response = self._grpc_client.services.repair_tools.find_stitch_faces(
            faces=body_ids
        )
        parent_design = get_design_from_body(bodies[0])
        return [
            StitchFaceProblemAreas(
                f"{res['id']}",
                self._grpc_client,
                get_bodies_from_ids(parent_design, res["bodies"]),
            )
            for res in problem_areas_response["problems"]
        ]

    @protect_grpc
    @min_backend_version(25, 2, 0)
    def find_simplify(self, bodies: list["Body"]) -> list[UnsimplifiedFaceProblemAreas]:
        """Detect faces in a body that can be simplified.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies to search.

        Returns
        -------
        list[UnsimplifiedFaceProblemAreas]
            List of objects representing unsimplified face problem areas.
        """
        from ansys.geometry.core.designer.body import Body

        check_type_all_elements_in_iterable(bodies, Body)
        body_ids = [body.id for body in bodies]

        parent_design = get_design_from_body(bodies[0])
        problem_areas_response = self._grpc_client.services.repair_tools.find_simplify(
            selection=body_ids
        )

        return [
            UnsimplifiedFaceProblemAreas(
                f"{res['id']}",
                self._grpc_client,
                get_faces_from_ids(parent_design, res["bodies"]),
            )
            for res in problem_areas_response["problems"]
        ]

    @protect_grpc
    @min_backend_version(25, 2, 0)
    def find_interferences(
        self, bodies: list["Body"], cut_smaller_body: bool = False
    ) -> list[InterferenceProblemAreas]:
        """Find the interference problem areas.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that small faces are investigated on.
        cut_smaller_body : bool, optional
            Whether to cut the smaller body if an intererference is found.
            By default, False.

        Returns
        -------
        list[InterfenceProblemAreas]
            List of objects representing interference problem areas.

        Notes
        -----
        This method finds and returns a list of ids of interference problem areas
        objects.
        """
        from ansys.geometry.core.designer.body import Body

        if not bodies:
            return []

        # Verify inputs
        check_type_all_elements_in_iterable(bodies, Body)
        check_type(cut_smaller_body, bool)

        parent_design = get_design_from_body(bodies[0])
        body_ids = [body.id for body in bodies]
        # cut_smaller_body_bool = BoolValue(value=cut_smaller_body)
        problem_areas_response = self._grpc_client.services.repair_tools.find_interferences(
            bodies=body_ids, cut_smaller_body=cut_smaller_body
        )

        return [
            InterferenceProblemAreas(
                f"{res['id']}",
                self._grpc_client,
                get_bodies_from_ids(parent_design, res["bodies"]),
            )
            for res in problem_areas_response["problems"]
        ]

    @protect_grpc
    @min_backend_version(25, 2, 0)
    def find_and_fix_short_edges(
        self, bodies: list["Body"], length: Real = 0.0, comprehensive_result: bool = False
    ) -> RepairToolMessage:
        """Find and fix the short edge problem areas.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that short edges are investigated on.
        length : Real, optional
            The maximum length of the edges. By default, 0.0.
        comprehensive_result : bool, optional
            Whether to fix all problem areas individually.
            By default, False.

        Returns
        -------
        RepairToolMessage
            Message containing number of problem areas found/fixed, created and/or modified bodies.

        Notes
        -----
        This method finds the short edges in the bodies and fixes them.
        """
        from ansys.geometry.core.designer.body import Body

        check_type_all_elements_in_iterable(bodies, Body)
        check_type(length, Real)
        check_type(comprehensive_result, bool)

        if not bodies:
            return RepairToolMessage(False, [], [], 0, 0)

        body_ids = [body.id for body in bodies]

        response = self._grpc_client.services.repair_tools.find_and_fix_short_edges(
            selection=body_ids,
            length=length,
            comprehensive_result=comprehensive_result,
        )

        parent_design = get_design_from_body(bodies[0])
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            success=response["success"],
            found=response["found"],
            repaired=response["repaired"],
            created_bodies=[],
            modified_bodies=[],
        )
        return message

    @protect_grpc
    @min_backend_version(25, 2, 0)
    def find_and_fix_extra_edges(
        self, bodies: list["Body"], comprehensive_result: bool = False
    ) -> RepairToolMessage:
        """Find and fix the extra edge problem areas.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that short edges are investigated on.
        length : Real
            The maximum length of the edges.
        comprehensive_result : bool, optional
            Whether to fix all problem areas individually.
            By default, False.

        Returns
        -------
        RepairToolMessage
            Message containing number of problem areas found/fixed, created and/or modified bodies.

        Notes
        -----
        This method finds the extra edges in the bodies and fixes them.
        """
        from ansys.geometry.core.designer.body import Body

        check_type_all_elements_in_iterable(bodies, Body)
        check_type(comprehensive_result, bool)

        if not bodies:
            return RepairToolMessage(False, [], [], 0, 0)

        body_ids = [body.id for body in bodies]
        response = self._grpc_client.services.repair_tools.find_and_fix_extra_edges(
            selection=body_ids,
            comprehensive_result=comprehensive_result,
        )

        parent_design = get_design_from_body(bodies[0])
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            response["success"],
            response["created_bodies_monikers"],
            response["modified_bodies_monikers"],
            response["found"],
            response["repaired"],
        )
        return message

    @protect_grpc
    @min_backend_version(25, 2, 0)
    def find_and_fix_split_edges(
        self,
        bodies: list["Body"],
        angle: Real = 0.0,
        length: Real = 0.0,
        comprehensive_result: bool = False,
    ) -> RepairToolMessage:
        """Find and fix the split edge problem areas.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies that split edges are investigated on.
        angle : Real, optional
            The maximum angle between edges. By default, 0.0.
        length : Real, optional
            The maximum length of the edges. By default, 0.0.
        comprehensive_result : bool, optional
            Whether to fix all problem areas individually.
            By default, False.

        Returns
        -------
        RepairToolMessage
            Message containing number of problem areas found/fixed, created and/or modified bodies.

        Notes
        -----
        This method finds the extra edges in the bodies and fixes them.
        """
        from ansys.geometry.core.designer.body import Body

        check_type_all_elements_in_iterable(bodies, Body)
        check_type(angle, Real)
        check_type(length, Real)
        check_type(comprehensive_result, bool)

        if not bodies:
            return RepairToolMessage(False, [], [], 0, 0)

        body_ids = [body.id for body in bodies]

        response = self._grpc_client.services.repair_tools.find_and_fix_split_edges(
            bodies_or_faces=body_ids,
            angle=angle,
            length=length,
            comprehensive_result=comprehensive_result,
        )

        parent_design = get_design_from_body(bodies[0])
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            response["success"],
            response["created_bodies_monikers"],
            response["modified_bodies_monikers"],
            response["found"],
            response["repaired"],
        )
        return message

    @protect_grpc
    @min_backend_version(25, 2, 0)
    def find_and_fix_simplify(
        self, bodies: list["Body"], comprehensive_result: bool = False
    ) -> RepairToolMessage:
        """Find and simplify the provided geometry.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies to be simplified.
        comprehensive_result : bool, optional
            Whether to fix all problem areas individually.
            By default, False.

        Returns
        -------
        RepairToolMessage
            Message containing number of problem areas found/fixed, created and/or modified bodies.

        Notes
        -----
        This method simplifies the provided geometry.
        """
        from ansys.geometry.core.designer.body import Body

        check_type_all_elements_in_iterable(bodies, Body)
        check_type(comprehensive_result, bool)

        if not bodies:
            return RepairToolMessage(False, [], [], 0, 0)

        body_ids = [body.id for body in bodies]

        response = self._grpc_client.services.repair_tools.find_and_fix_simplify(
            selection=body_ids,
            comprehensive_result=comprehensive_result,
        )

        parent_design = get_design_from_body(bodies[0])
        parent_design._update_design_inplace()
        message = RepairToolMessage(
            response["success"],
            response["created_bodies_monikers"],
            response["modified_bodies_monikers"],
            response["found"],
            response["repaired"],
        )
        return message

    def inspect_geometry(self, bodies: list["Body"] = None) -> list[InspectResult]:
        """Return a list of geometry issues organized by body.

        This method inspects the geometry and returns a list of the issues grouped by
        the body where they are found.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies to inspect the geometry for.
            All bodies are inspected if the argument is not given.

        Returns
        -------
        list[IssuesByBody]
            List of objects representing geometry issues and the bodies where issues are found.
        """
        parent_design = self._modeler.get_active_design()
        body_ids = [] if bodies is None else [body._grpc_id for body in bodies]
        inspect_result_response_dict = self._grpc_client.services.repair_tools.inspect_geometry(
            parent_design=parent_design, bodies=body_ids
        )
        return self.__create_inspect_result_from_response(
            parent_design, inspect_result_response_dict["issues_by_body"]
        )

    def __create_inspect_result_from_response(
        self, design, inspect_geometry_results: list[dict]
    ) -> list[InspectResult]:
        inspect_results = []
        for inspect_geometry_result in inspect_geometry_results:
            body = get_bodies_from_ids(design, [inspect_geometry_result["body"]["id"]])
            issues = self.__create_issues_from_response(inspect_geometry_result["issues"])
            inspect_result = InspectResult(
                grpc_client=self._grpc_client, body=body[0], issues=issues
            )
            inspect_results.append(inspect_result)

        return inspect_results

    def __create_issues_from_response(
        self,
        inspect_geometry_result_issues: list[dict],
    ) -> list[GeometryIssue]:
        issues = []
        for issue in inspect_geometry_result_issues:
            message_type = issue["message_type"]
            message_id = issue["message_id"]
            message = issue["message"]

            faces = [face["id"] for face in issue.get("faces", [])]
            edges = [edge["id"] for edge in issue.get("edges", [])]

            geometry_issue = GeometryIssue(
                message_type=message_type,
                message_id=message_id,
                message=message,
                faces=faces,
                edges=edges,
            )
            issues.append(geometry_issue)
        return issues

    @protect_grpc
    @min_backend_version(25, 2, 0)
    def repair_geometry(self, bodies: list["Body"] = None) -> RepairToolMessage:
        """Attempt to repair the geometry for the given bodies.

        This method inspects the geometry for the given bodies and attempts to repair them.

        Parameters
        ----------
        bodies : list[Body]
            List of bodies where to attempt to repair the geometry.
            All bodies are repaired if the argument is not given.

        Returns
        -------
        RepairToolMessage
            Message containing success of the operation.
        """
        body_ids = [] if bodies is None else [body._grpc_id for body in bodies]
        repair_result_response = self._grpc_client.services.repair_tools.repair_geometry(
            bodies=body_ids
        )

        message = RepairToolMessage(repair_result_response["success"], [], [])
        return message



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for repair tool message."""


class RepairToolMessage:
    """Provides return message for the repair tool methods."""

    def __init__(
        self,
        success: bool,
        created_bodies: list[str],
        modified_bodies: list[str],
        found: int = -1,
        repaired: int = -1,
    ):
        """Initialize a new instance of the extra edge problem area class.

        Parameters
        ----------
        success: bool
            True if the repair operation was effective, false if it is not.
        created_bodies: list[str]
            List of bodies created after the repair operation.
        modified_bodies: list[str]
            List of bodies modified after the repair operation.
        found: int, default: -1
            Number of problem areas found for the repair operation.
            If default, the operation does not provide the number of found problem areas.
        repaired: int, default: -1
            Number of problem areas repaired during the repair operation.
            If default, the operation does not provide the number of fixed problem areas.


        """
        self._success = success
        self._created_bodies = created_bodies
        self._modified_bodies = modified_bodies
        self._found = found
        self._repaired = repaired

    @property
    def success(self) -> bool:
        """The success of the repair operation."""
        return self._success

    @property
    def created_bodies(self) -> list[str]:
        """The list of the created bodies after the repair operation."""
        return self._created_bodies

    @property
    def modified_bodies(self) -> list[str]:
        """The list of the modified bodies after the repair operation."""
        return self._modified_bodies

    @property
    def found(self) -> int:
        """Number of problem areas found for the repair operation."""
        return self._found

    @property
    def repaired(self) -> int:
        """Number of problem areas repaired during the repair operation."""
        return self._repaired



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing a wrapped abstraction of the gRPC stubs."""

import atexit
import logging
from pathlib import Path
import time
from typing import Optional

from beartype import beartype as check_input_types
import grpc
from grpc._channel import _InactiveRpcError
from grpc_health.v1 import health_pb2, health_pb2_grpc
import semver

from ansys.geometry.core._grpc._services._service import _GRPCServices
from ansys.geometry.core.connection.backend import BackendType
import ansys.geometry.core.connection.defaults as pygeom_defaults
from ansys.geometry.core.connection.docker_instance import LocalDockerInstance
from ansys.geometry.core.connection.product_instance import ProductInstance
from ansys.geometry.core.logger import LOG, PyGeometryCustomAdapter
from ansys.geometry.core.misc.checks import deprecated_method
from ansys.geometry.core.typing import Real

try:
    from ansys.platform.instancemanagement import Instance
except ModuleNotFoundError:  # pragma: no cover
    pass


def _create_geometry_channel(target: str) -> grpc.Channel:
    """Create a Geometry service gRPC channel.

    Parameters
    ----------
    target : str
        Target of the channel. This is usually a string in the form of
        ``host:port``.

    Returns
    -------
    ~grpc.Channel
        gRPC channel for the Geometry service.

    Notes
    -----
    Contains specific options for the Geometry service.
    """
    return grpc.insecure_channel(
        target,
        options=[
            ("grpc.max_receive_message_length", pygeom_defaults.MAX_MESSAGE_LENGTH),
            ("grpc.max_send_message_length", pygeom_defaults.MAX_MESSAGE_LENGTH),
        ],
    )


def wait_until_healthy(channel: grpc.Channel | str, timeout: float) -> grpc.Channel:
    """Wait until a channel is healthy before returning.

    Parameters
    ----------
    channel : ~grpc.Channel | str
        Channel that must be established and healthy. The target can also be
        passed in. In that case, a channel is created using the default insecure channel.
    timeout : float
        Timeout in seconds. Attempts are made with the following backoff strategy:

        * Starts with 0.1 seconds.
        * If the attempt fails, double the timeout.
        * This is repeated until the next timeoff exceeds the
          value for the remaining time. In that case, a final attempt
          is made with the remaining time.
        * If the total elapsed time exceeds the value for the ``timeout`` parameter,
          a ``TimeoutError`` is raised.

    Returns
    -------
    grpc.Channel
        The channel that was passed in. This channel is guaranteed to be healthy.
        If a string was passed in, a channel is created using the default insecure channel.

    Raises
    ------
    TimeoutError
        Raised when the total elapsed time exceeds the value for the ``timeout`` parameter.
    """
    t_max = time.time() + timeout
    t_out = 0.1

    # If the channel is a string, create a channel using the default insecure channel
    channel_creation_required = True if isinstance(channel, str) else False
    tmp_channel = None

    while time.time() < t_max:
        try:
            tmp_channel = (
                _create_geometry_channel(channel) if channel_creation_required else channel
            )
            health_stub = health_pb2_grpc.HealthStub(tmp_channel)
            request = health_pb2.HealthCheckRequest(service="")

            out = health_stub.Check(request, timeout=t_out)
            if out.status is health_pb2.HealthCheckResponse.SERVING:
                break
        except _InactiveRpcError:
            # Duplicate timeout and try again
            t_now = time.time()
            t_out *= 2
            # If we have time to try again, continue.. but if we don't,
            # just try for the remaining time
            if t_now + t_out > t_max:
                t_out = t_max - t_now
            continue
    else:
        target_str = tmp_channel._channel.target().decode()
        raise TimeoutError(
            f"Channel health check to target '{target_str}' timed out after {timeout} seconds."
        )

    return tmp_channel


class GrpcClient:
    """Wraps the gRPC connection for the Geometry service.

    Parameters
    ----------
    host : str, default: DEFAULT_HOST
        Host where the server is running.
    port : str or int, default: DEFAULT_PORT
        Port number where the server is running.
    channel : ~grpc.Channel, default: None
        gRPC channel for server communication.
    remote_instance : ansys.platform.instancemanagement.Instance, default: None
        Corresponding remote instance when the Geometry service
        is launched through `PyPIM <https://github.com/ansys/pypim>`_.
        This instance is deleted when calling the
        :func:`GrpcClient.close <ansys.geometry.core.client.GrpcClient.close >`
        method.
    docker_instance : LocalDockerInstance, default: None
        Corresponding local Docker instance when the Geometry service is launched using
        the ``launch_docker_modeler()`` method. This local Docker instance is deleted
        when the :func:`GrpcClient.close <ansys.geometry.core.client.GrpcClient.close >`
        method is called.
    product_instance : ProductInstance, default: None
        Corresponding local product instance when the product (Discovery or SpaceClaim)
        is launched through the ``launch_modeler_with_geometry_service()``,
        ``launch_modeler_with_discovery()`` or the ``launch_modeler_with_spaceclaim()``
        interface. This instance will be deleted
        when the :func:`GrpcClient.close <ansys.geometry.core.client.GrpcClient.close >`
        method is called.
    timeout : real, default: 120
        Maximum time to spend trying to make the connection.
    logging_level : int, default: INFO
        Logging level to apply to the client.
    logging_file : str or Path, default: None
        File to output the log to, if requested.
    proto_version: str or None, default: None
        Version of the gRPC protocol to use. If ``None``, the latest version is used.
    """

    @check_input_types
    def __init__(
        self,
        host: str = pygeom_defaults.DEFAULT_HOST,
        port: str | int = pygeom_defaults.DEFAULT_PORT,
        channel: grpc.Channel | None = None,
        remote_instance: Optional["Instance"] = None,
        docker_instance: LocalDockerInstance | None = None,
        product_instance: ProductInstance | None = None,
        timeout: Real = 120,
        logging_level: int = logging.INFO,
        logging_file: Path | str | None = None,
        proto_version: str | None = None,
    ):
        """Initialize the ``GrpcClient`` object."""
        self._closed = False
        self._remote_instance = remote_instance
        self._docker_instance = docker_instance
        self._product_instance = product_instance
        self._grpc_health_timeout = timeout

        if channel:
            # Used for PyPIM when directly providing a channel
            self._target = str(channel)
            self._channel = wait_until_healthy(channel, self._grpc_health_timeout)
        else:
            self._target = f"{host}:{port}"
            self._channel = wait_until_healthy(self._target, self._grpc_health_timeout)

        # Initialize the gRPC services
        self._services = _GRPCServices(self._channel, version=proto_version)

        # Once connection with the client is established, create a logger
        self._log = LOG.add_instance_logger(
            name=self._target, client_instance=self, level=logging_level
        )
        if logging_file:
            if isinstance(logging_file, Path):
                logging_file = str(logging_file)
            self._log.log_to_file(filename=logging_file, level=logging_level)

        # Retrieve the backend information
        response = self._services.admin.get_backend()

        # Store the backend type and version
        self._backend_type = response.get("backend")
        self._backend_version = response.get("version")

        # Register the close method to be called at exit - irrespectively of
        # the user calling it or not...
        atexit.register(self.close)

    @property
    def backend_type(self) -> BackendType:
        """Backend type.

        Options are ``Windows Service``, ``Linux Service``, ``Discovery``,
        and ``SpaceClaim``.

        Notes
        -----
        This method might return ``None`` because determining the backend type is
        not straightforward.
        """
        return self._backend_type

    @property
    def backend_version(self) -> semver.version.Version:
        """Get the current backend version.

        Returns
        -------
        ~semver.version.Version
            Backend version.
        """
        return self._backend_version

    @property
    @deprecated_method(
        info="Multiple designs for the same service are no longer supported.",
        version="0.9.0",
        remove="0.11.0",
    )
    def multiple_designs_allowed(self) -> bool:
        """Flag indicating whether multiple designs are allowed.

        Notes
        -----
        Currently, only one design is allowed per service. This method will always
        return ``False``.
        """
        return False

    @property
    def channel(self) -> grpc.Channel:
        """Client gRPC channel."""
        return self._channel

    @property
    def services(self) -> _GRPCServices:
        """GRPC services."""
        return self._services

    @property
    def log(self) -> PyGeometryCustomAdapter:
        """Specific instance logger."""
        return self._log

    @property
    def is_closed(self) -> bool:
        """Flag indicating whether the client connection is closed."""
        return self._closed

    @property
    def healthy(self) -> bool:
        """Flag indicating whether the client channel is healthy."""
        if self._closed:
            return False
        try:
            return self.services.admin.get_service_status().get("healthy")
        except Exception:  # pragma: no cover
            return False

    def __repr__(self) -> str:
        """Represent the client as a string."""
        lines = []
        lines.append(f"Ansys Geometry Modeler Client ({hex(id(self))})")
        lines.append(f"  Target:     {self._target}")
        if self._closed:
            lines.append("  Connection: Closed")
        elif self.healthy:
            lines.append("  Connection: Healthy")
        else:
            lines.append("  Connection: Unhealthy")  # pragma: no cover
        return "\n".join(lines)

    def close(self):
        """Close the channel.

        Notes
        -----
        If an instance of the Geometry service was started using
        `PyPIM <https://github.com/ansys/pypim>`_, this instance is
        deleted. Furthermore, if a local Docker instance
        of the Geometry service was started, it is stopped.
        """
        if self._closed is True:  # pragma: no cover
            self.log.debug("Connection is already closed. Ignoring request.")
            return

        if self._remote_instance:
            self._remote_instance.delete()  # pragma: no cover
        elif self._docker_instance:
            if not self._docker_instance.existed_previously:
                self._docker_instance.container.stop()
            else:
                self.log.warning(
                    "Geometry service was not shut down because it was already running..."
                )
        elif self._product_instance:
            self._product_instance.close()

        self._closed = True
        self._channel.close()

    def target(self) -> str:
        """Get the target of the channel."""
        if self._closed:
            return ""
        return self._channel._channel.target().decode()

    def get_name(self) -> str:
        """Get the target name of the connection."""
        return self._target

    @check_input_types
    def _get_service_logs(
        self,
        all_logs: bool = False,
        dump_to_file: bool = False,
        logs_folder: str | Path | None = None,
    ) -> str | dict[str, str] | Path:
        """Get the service logs.

        Parameters
        ----------
        all_logs : bool, default: False
            Flag indicating whether all logs should be retrieved. By default,
            only the current logs are retrieved.
        dump_to_file : bool, default: False
            Flag indicating whether the logs should be dumped to a file.
            By default, the logs are not dumped to a file.
        logs_folder : str,  Path or None, default: None
            Name of the folder where the logs should be dumped. This parameter
            is only used if the ``dump_to_file`` parameter is set to ``True``.

        Returns
        -------
        str
            Service logs as a string. This is returned if the ``dump_to_file`` parameter
            is set to ``False``.
        dict[str, str]
            Dictionary containing the logs. The keys are the logs names,
            and the values are the logs as strings. This is returned if the ``all_logs``
            parameter is set to ``True`` and the ``dump_to_file`` parameter
            is set to ``False``.
        Path
            Path to the folder containing the logs (if the ``all_logs``
            parameter is set to ``True``) or the path to the log file (if only
            the current logs are retrieved). The ``dump_to_file`` parameter
            must be set to ``True``.
        """
        logs: dict[str, str] = self._services.admin.get_logs(all_logs=all_logs).get("logs")

        # Let's handle the various scenarios...
        if not dump_to_file:
            return logs if all_logs else next(iter(logs.values()))
        else:
            if logs_folder is None:
                logs_folder = Path.cwd()
            elif isinstance(logs_folder, str):
                logs_folder = Path(logs_folder)

            logs_folder.mkdir(parents=True, exist_ok=True)
            for log_name, log_content in logs.items():
                with (logs_folder / log_name).open("w") as f:
                    f.write(log_content)

            return (logs_folder / log_name) if len(logs) == 1 else logs_folder



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides PyAnsys Geometry-specific errors."""

from functools import wraps
import signal
import threading

from grpc._channel import _InactiveRpcError, _MultiThreadedRendezvous

from ansys.geometry.core.logger import LOG

SIGINT_TRACKER = []


class GeometryRuntimeError(RuntimeError):
    """Provides error message when Geometry service passes a runtime error."""

    pass


class GeometryExitedError(RuntimeError):
    """Provides error message to raise when Geometry service has exited.

    Parameters
    ----------
    msg : str, default: "Geometry service has exited."
        Message to raise.
    """

    def __init__(self, msg="Geometry service has exited."):
        """Initialize the ``GeometryExitedError`` error."""
        RuntimeError.__init__(self, msg)  # pragma: no cover


# handler for protect_grpc
def handler(sig, frame):  # pragma: no cover
    """Pass signal to the custom interrupt handler."""
    LOG.info("KeyboardInterrupt received. Waiting until Geometry service execution finishes.")
    SIGINT_TRACKER.append(True)


def protect_grpc(func):
    """Capture gRPC exceptions and raise a more succinct error message.

    This method captures the ``KeyboardInterrupt`` exception to avoid
    segfaulting the Geometry service.

    While this works some of the time, it does not work all of the time. For some
    reason, gRPC still captures SIGINT.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        """Capture gRPC exceptions and KeyboardInterrupt errors.

        Returns
        -------
        out
            Result of the function wrapped.

        Raises
        ------
        GeometryExitedError
            If a gRPC error of type InactiveRpcError, MultiThreadedRendezvous is observed.
        KeyboardInterrupt
            If a KeyboardInterrupt error is observed.
        """
        # capture KeyboardInterrupt
        old_handler = None
        if threading.current_thread().__class__.__name__ == "_MainThread":
            if threading.current_thread().is_alive():
                old_handler = signal.signal(signal.SIGINT, handler)

        # Capture gRPC exceptions
        try:
            out = func(*args, **kwargs)
        except (_InactiveRpcError, _MultiThreadedRendezvous) as error:  # pragma: no cover
            raise GeometryExitedError(
                f"Geometry service connection terminated: {error.details()}"
            ) from None

        if threading.current_thread().__class__.__name__ == "_MainThread":
            received_interrupt = bool(SIGINT_TRACKER)

            # always clear and revert to old handler
            SIGINT_TRACKER.clear()
            if old_handler:
                signal.signal(signal.SIGINT, old_handler)

            if received_interrupt:  # pragma: no cover
                raise KeyboardInterrupt("Interrupted during Geometry service execution")

        return out

    return wrapper



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the repair tools service implementation (abstraction layer).

This module defines an abstract base class for a gRPC-based repair tools service.
The class provides a set of abstract methods for identifying and repairing various
geometry issues, such as split edges, extra edges, duplicate faces, and more.
"""

from abc import ABC, abstractmethod

import grpc


class GRPCRepairToolsService(ABC):  # pragma: no cover
    """Abstract base class for gRPC-based repair tools service.

    Parameters
    ----------
    channel: grpc.Channel
        The gRPC channel used to communicate with the service.
    """

    def __init__(self, channel: grpc.Channel):
        """Initialize the gRPC repair tools service."""

    @abstractmethod
    def find_split_edges(self, **kwargs) -> dict:
        """Identify split edges in the geometry."""
        pass

    @abstractmethod
    def find_extra_edges(self, **kwargs) -> dict:
        """Identify extra edges in the geometry."""
        pass

    @abstractmethod
    def find_inexact_edges(self, **kwargs) -> dict:
        """Identify inexact edges in the geometry."""
        pass

    @abstractmethod
    def find_short_edges(self, **kwargs) -> dict:
        """Identify short edges in the geometry."""
        pass

    @abstractmethod
    def find_duplicate_faces(self, **kwargs) -> dict:
        """Identify duplicate faces in the geometry."""
        pass

    @abstractmethod
    def find_missing_faces(self, **kwargs) -> dict:
        """Identify missing faces in the geometry."""
        pass

    @abstractmethod
    def find_small_faces(self, **kwargs) -> dict:
        """Identify small faces in the geometry."""
        pass

    @abstractmethod
    def find_stitch_faces(self, **kwargs) -> dict:
        """Identify faces that can be stitched together in the geometry."""
        pass

    @abstractmethod
    def find_simplify(self, **kwargs) -> dict:
        """Identify areas in the geometry that can be simplified."""
        pass

    @abstractmethod
    def find_interferences(self, **kwargs) -> dict:
        """Identify interferences in the geometry."""
        pass

    @abstractmethod
    def find_and_fix_short_edges(self, **kwargs) -> dict:
        """Identify and fix short edges in the geometry."""
        pass

    @abstractmethod
    def find_and_fix_extra_edges(self, **kwargs) -> dict:
        """Identify and fix extra edges in the geometry."""
        pass

    @abstractmethod
    def find_and_fix_split_edges(self, **kwargs) -> dict:
        """Identify and fix split edges in the geometry."""
        pass

    @abstractmethod
    def find_and_fix_simplify(self, **kwargs) -> dict:
        """Identify and simplify areas in the geometry."""
        pass

    @abstractmethod
    def inspect_geometry(self, **kwargs) -> dict:
        """Inspect the geometry for issues."""
        pass

    @abstractmethod
    def repair_geometry(self, **kwargs) -> dict:
        """Repair the geometry by addressing identified issues."""
        pass



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides plotting for various PyAnsys Geometry objects."""

# First, verify graphics are available
try:
    from ansys.geometry.core.misc.checks import run_if_graphics_required

    run_if_graphics_required()
except ImportError as err:  # pragma: no cover
    raise err

from itertools import cycle
from pathlib import Path
from typing import Any

import numpy as np
from pygltflib.utils import gltf2glb
import pyvista as pv
from pyvista.plotting.tools import create_axes_marker

import ansys.geometry.core as pyansys_geometry
from ansys.geometry.core.designer.body import Body, MasterBody
from ansys.geometry.core.designer.component import Component
from ansys.geometry.core.designer.design import Design
from ansys.geometry.core.designer.designpoint import DesignPoint
from ansys.geometry.core.designer.face import Face
from ansys.geometry.core.errors import GeometryRuntimeError
from ansys.geometry.core.logger import LOG
from ansys.geometry.core.math.frame import Frame
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.misc.auxiliary import DEFAULT_COLOR
from ansys.geometry.core.plotting.widgets import ShowDesignPoints
from ansys.geometry.core.sketch.sketch import Sketch
from ansys.tools.visualization_interface import (
    Color,
    EdgePlot,
    MeshObjectPlot,
    Plotter as PlotterInterface,
)
from ansys.tools.visualization_interface.backends.pyvista import PyVistaBackend

POLYDATA_COLOR_CYCLER = cycle(pv.colors.get_cycler("matplotlib"))


class GeometryPlotter(PlotterInterface):
    """Plotter for PyAnsys Geometry objects.

    This class is an implementation of the PlotterInterface class.

    Parameters
    ----------
    use_trame : bool, optional
        Whether to use trame visualizer or not, by default None.
    use_service_colors : bool, optional
        Whether to use service colors or not, by default None.
    allow_picking : bool, optional
        Whether to allow picking or not, by default False.
    show_plane : bool, optional
        Whether to show the plane in the scene, by default True.
    """

    def __init__(
        self,
        use_trame: bool | None = None,
        use_service_colors: bool | None = None,
        allow_picking: bool = False,
        show_plane: bool = True,
    ) -> None:
        """Initialize the GeometryPlotter class."""
        self._backend = PyVistaBackend(use_trame=use_trame, allow_picking=allow_picking)
        super().__init__(backend=self._backend)

        self._backend._allow_picking = allow_picking
        self._backend._use_trame = use_trame
        self._backend.add_widget(ShowDesignPoints(self))
        self._backend._pl._show_plane = show_plane

        # Store the use_service_colors flag
        self._use_service_colors = (
            use_service_colors
            if use_service_colors is not None
            else pyansys_geometry.USE_SERVICE_COLORS
        )

    @property
    def use_service_colors(self) -> bool:
        """Indicates whether to use service colors for plotting purposes."""
        return self._use_service_colors

    def add_frame(self, frame: Frame, plotting_options: dict | None = None) -> None:
        """Plot a frame in the scene.

        Parameters
        ----------
        frame : Frame
            Frame to render in the scene.
        plotting_options : dict, default: None
            dictionary containing parameters accepted by the
            :func:`pyvista.create_axes_marker` class for customizing
            the frame rendering in the scene.
        """
        # Use default plotting options if required
        if plotting_options is None:
            plotting_options = dict(labels_off=True, cone_radius=0.2)

        # Create the axes actor
        axes = create_axes_marker(**plotting_options)

        # Transpose the matrix for fixing rotation sense in VTK
        arr = np.vstack((frame.global_to_local_rotation, frame.origin)).T
        arr = np.vstack((arr, [0, 0, 0, 1]))

        # Apply matrix transformation to the actor
        axes.SetUserMatrix(pv.vtkmatrix_from_array(arr))

        # Render the actor in the scene
        self._backend.pv_interface.scene.add_actor(axes)

    def add_plane(
        self,
        plane: Plane,
        plane_options: dict | None = None,
        plotting_options: dict | None = None,
    ) -> None:
        """Plot a plane in the scene.

        Parameters
        ----------
        plane : Plane
            Plane to render in the scene.
        plane_options : dict, default: None
            dictionary containing parameters accepted by the
            :func:`pyvista.Plane  <pyvista.Plane>` function for customizing the mesh
            representing the plane.
        plotting_options : dict, default: None
            dictionary containing parameters accepted by the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method for
            customizing the mesh rendering of the plane.
        """
        # Impose default plane options if none provided
        if plane_options is None:
            plane_options = dict(i_size=10, j_size=10)

        plane_mesh = pv.Plane(
            center=plane.origin.tolist(), direction=plane.direction_z.tolist(), **plane_options
        )
        plane_meshobj = MeshObjectPlot(custom_object=plane, mesh=plane_mesh)
        # Impose default plotting options if none provided
        if plotting_options is None:
            plotting_options = dict(color="blue", opacity=0.1)

        self._backend.pv_interface.plot(plane_meshobj, **plotting_options)

    def add_sketch(
        self,
        sketch: Sketch,
        show_plane: bool = False,
        show_frame: bool = False,
        **plotting_options: dict | None,
    ) -> None:
        """Plot a sketch in the scene.

        Parameters
        ----------
        sketch : Sketch
            Sketch to render in the scene.
        show_plane : bool, default: False
            Whether to render the sketch plane in the scene.
        show_frame : bool, default: False
            Whether to show the frame in the scene.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        # Show the sketch plane if required
        if show_plane:
            self.add_plane(sketch._plane)

        # Show the sketch plane if required
        if show_frame:
            self.add_frame(sketch._plane)

        if "clipping_plane" in plotting_options:
            LOG.warning("Clipping is not available in Sketch objects.")
            plotting_options.pop("clipping_plane")

        self.add_sketch_polydata(
            sketch.sketch_polydata_faces(), sketch, opacity=0.7, **plotting_options
        )
        self.add_sketch_polydata(sketch.sketch_polydata_edges(), sketch, **plotting_options)

    def add_body_edges(self, body_plot: MeshObjectPlot, **plotting_options: dict | None) -> None:
        """Add the outer edges of a body to the plot.

        This method has the side effect of adding the edges to the GeomObject that
        you pass through the parameters.

        Parameters
        ----------
        body_plot : MeshObjectPlot
            Body of which to add the edges.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        edge_plot_list = []
        for edge in body_plot.custom_object.edges:
            line = pv.Line(edge.start, edge.end)
            edge_actor = self._backend._pl.scene.add_mesh(
                line, line_width=10, color=Color.EDGE.value, **plotting_options
            )
            edge_actor.SetVisibility(False)
            edge_plot = EdgePlot(edge_actor, edge, body_plot)
            edge_plot_list.append(edge_plot)
        body_plot.edges = edge_plot_list

    def add_body(self, body: Body, merge: bool = False, **plotting_options: dict | None) -> None:
        """Add a body to the scene.

        Parameters
        ----------
        body : Body
            Body to add.
        merge : bool, default: False
            Whether to merge the body into a single mesh. When ``True``, the
            individual faces of the tessellation are merged. This
            preserves the number of triangles and only merges the topology.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments,
            see the :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        try:
            if body.is_suppressed:
                return
        except GeometryRuntimeError:  # pragma: no cover
            # For backward compatibility with older versions of PyAnsys Geometry
            # Inserted in 25R2
            pass

        if self.use_service_colors:
            faces = body.faces
            body_color = body.color
            if not merge:
                for face in faces:
                    face_color = face.color
                    if face_color == DEFAULT_COLOR or face_color[0:7] == DEFAULT_COLOR.lower():
                        plotting_options["color"] = body_color
                        plotting_options["opacity"] = body.opacity
                    else:
                        plotting_options["color"] = face_color
                        plotting_options["opacity"] = face.opacity
                    self._backend.pv_interface.plot(face.tessellate(), **plotting_options)
                return
            else:
                dataset = body.tessellate(merge=True)
                plotting_options["color"] = body_color
                plotting_options["opacity"] = body.opacity
                self._backend.pv_interface.plot(dataset, **plotting_options)
                return
        # WORKAROUND: multi_colors is not properly supported in PyVista PolyData
        # so if multi_colors is True and merge is True (returns PolyData) then
        # we need to set the color manually
        elif (
            merge
            and plotting_options.get("multi_colors", False)
            and "color" not in plotting_options
        ):
            plotting_options["color"] = next(POLYDATA_COLOR_CYCLER)["color"]

        # Use the default PyAnsys Geometry add_mesh arguments
        self._backend.pv_interface.set_add_mesh_defaults(plotting_options)
        dataset = body.tessellate(merge=merge)
        body_plot = MeshObjectPlot(custom_object=body, mesh=dataset)
        self._backend.pv_interface.plot(body_plot, **plotting_options)

        # Edges should ONLY be plotted individually if picking is enabled
        if self._backend._allow_picking:
            self.add_body_edges(body_plot)

    def add_face(self, face: Face, **plotting_options: dict | None) -> None:
        """Add a face to the scene.

        Parameters
        ----------
        face : Face
            Face to add.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        self._backend.pv_interface.set_add_mesh_defaults(plotting_options)
        dataset = face.tessellate()
        if self.use_service_colors:
            face_color = face.color
            if face_color != DEFAULT_COLOR:
                plotting_options["color"] = face_color
                plotting_options["opacity"] = face.opacity
            else:
                plotting_options["color"] = face.body.color
                plotting_options["opacity"] = face.body.opacity
            self._backend.pv_interface.plot(dataset, **plotting_options)
            return

        # Otherwise...
        face_plot = MeshObjectPlot(custom_object=face, mesh=dataset)
        self._backend.pv_interface.plot(face_plot, **plotting_options)

    def add_component(
        self,
        component: Component,
        merge_component: bool = False,
        merge_bodies: bool = False,
        **plotting_options,
    ) -> None:
        """Add a component to the scene.

        Parameters
        ----------
        component : Component
            Component to add.
        merge_component : bool, default: False
            Whether to merge the component into a single dataset. When
            ``True``, all the individual bodies are effectively combined
            into a single dataset without any hierarchy.
        merge_bodies : bool, default: False
            Whether to merge each body into a single dataset. When ``True``,
            all the faces of each individual body are effectively combined
            into a single dataset without separating faces.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        if merge_component:
            self._backend.pv_interface.set_add_mesh_defaults(plotting_options)
            dataset = component.tessellate()
            component_polydata = MeshObjectPlot(component, dataset)
            self.plot(component_polydata, **plotting_options)
        else:
            self.add_component_by_body(
                component,
                merge_bodies=merge_bodies,
                **plotting_options,
            )

    def add_component_by_body(
        self, component: Component, merge_bodies: bool, **plotting_options: dict | None
    ) -> None:
        """Add a component on a per body basis.

        Parameters
        ----------
        component : Component
            Component to add.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.

        Notes
        -----
        This will allow to make use of the service colors. At the same time, it will be
        slower than the add_component method.
        """
        # Recursively add the bodies and components
        for body in component.bodies:
            self.add_body(body, merge=merge_bodies, **plotting_options)
        for comp in component.components:
            self.add_component_by_body(comp, merge_bodies=merge_bodies, **plotting_options)

    def add_sketch_polydata(
        self, polydata_entries: list[pv.PolyData], sketch: Sketch = None, **plotting_options
    ) -> None:
        """Add sketches to the scene from PyVista polydata.

        Parameters
        ----------
        polydata_entries : list[pyvista.PolyData]
            Polydata to add.
        sketch : Sketch, default: None
            Sketch to add.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        # Use the default PyAnsys Geometry add_mesh arguments
        mb = pv.MultiBlock()
        for polydata in polydata_entries:
            mb.append(polydata)

        if sketch is None:
            self.plot(mb, color=Color.EDGE.value, **plotting_options)
        else:
            sk_polydata = MeshObjectPlot(custom_object=sketch, mesh=mb)
            self.plot(sk_polydata, color=Color.EDGE.value, **plotting_options)

    def add_design_point(self, design_point: DesignPoint, **plotting_options) -> None:
        """Add a DesignPoint object to the plotter.

        Parameters
        ----------
        design_point : DesignPoint
            DesignPoint to add.
        """
        design_point = MeshObjectPlot(custom_object=design_point, mesh=design_point._to_polydata())

        # get the actor for the DesignPoint
        self._backend.pv_interface.plot(design_point, **plotting_options)

    def plot_iter(
        self,
        plotting_list: list[Any],
        name_filter: str = None,
        **plotting_options,
    ) -> None:
        """Add a list of any type of object to the scene.

        These types of objects are supported: ``Body``, ``Component``, ``list[pv.PolyData]``,
        ``pv.MultiBlock``, and ``Sketch``.

        Parameters
        ----------
        plotting_list : list[Any]
            list of objects you want to plot.
        name_filter : str, default: None
            Regular expression with the desired name or names you want to include in the plotter.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        for object in plotting_list:
            _ = self.plot(object, name_filter, **plotting_options)

    # Override add function from plotter
    def plot(self, plottable_object: Any, name_filter: str = None, **plotting_options) -> None:
        """Add a custom mesh to the plotter.

        Parameters
        ----------
        plottable_object : str, default: None
            Regular expression with the desired name or names you want to include in the plotter.
        name_filter: str, default: None
            Regular expression with the desired name or names you want to include in the plotter.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, depend of the backend implementation
            you are using.
        """
        # Check for custom parameters in plotting_options
        if "merge_bodies" in plotting_options:
            merge_bodies = plotting_options["merge_bodies"]
            plotting_options.pop("merge_bodies", None)
        else:
            merge_bodies = None

        if "merge_component" in plotting_options:
            merge_component = plotting_options["merge_component"]
            plotting_options.pop("merge_component", None)
        else:
            merge_component = None
        # Add the custom object to the plotter
        if isinstance(plottable_object, DesignPoint):
            self.add_design_point(plottable_object, **plotting_options)
        elif isinstance(plottable_object, Sketch):
            self.add_sketch(plottable_object, **plotting_options)
        elif isinstance(plottable_object, (Body, MasterBody)):
            self.add_body(plottable_object, merge_bodies, **plotting_options)
        elif isinstance(plottable_object, Face):
            self.add_face(plottable_object, **plotting_options)
        elif isinstance(plottable_object, (Design, Component)):
            self.add_component(plottable_object, merge_component, merge_bodies, **plotting_options)
        elif (
            isinstance(plottable_object, list)
            and len(plottable_object) > 0
            and all([isinstance(entry, pv.PolyData) for entry in plottable_object])
        ):
            self.add_sketch_polydata(plottable_object, **plotting_options)
        elif isinstance(plottable_object, list):
            self.plot_iter(plottable_object, name_filter, **plotting_options)
        elif isinstance(plottable_object, MeshObjectPlot):
            self._backend.pv_interface.set_add_mesh_defaults(plotting_options)
            self._backend.pv_interface.plot(plottable_object, name_filter, **plotting_options)
        else:
            # any left type should be a PyVista object
            self._backend.pv_interface.plot(plottable_object, name_filter, **plotting_options)

    def show(
        self,
        plotting_object: Any = None,
        screenshot: str | None = None,
        **plotting_options,
    ) -> None | list[Any]:
        """Show the plotter.

        Parameters
        ----------
        plotting_object : Any, default: None
            Object you can add to the plotter.
        screenshot : str, default: None
            Path to save a screenshot of the plotter.
        **plotting_options : dict, default: None
            Keyword arguments for the plotter. Arguments depend of the backend implementation
            you are using.
        """
        if plotting_object is not None:
            self.plot(plotting_object, **plotting_options)
        picked_objs = self._backend.show(screenshot=screenshot, **plotting_options)

        # Return the picked objects if picking is enabled... but as the actual PyAnsys
        # Geometry objects (or PyVista objects if not)
        if picked_objs:
            lib_objects = []
            for element in picked_objs:
                if isinstance(element, MeshObjectPlot):
                    lib_objects.append(element.custom_object)
                elif isinstance(element, EdgePlot):
                    lib_objects.append(element.edge_object)
                else:  # Either a PyAnsys Geometry object or a PyVista object
                    lib_objects.append(element)
            return lib_objects

    def export_glb(
        self, plotting_object: Any = None, filename: str | Path | None = None, **plotting_options
    ) -> Path:
        """Export the design to a glb file. Does not support picked objects.

        Parameters
        ----------
        plotting_object : Any, default: None
            Object you can add to the plotter.
        filename : str | ~pathlib.Path, default: None
            Path to save a GLB file of the plotter. If None, the file will be saved as
            temp_glb.glb.
        **plotting_options : dict, default: None
            Keyword arguments for the plotter. Arguments depend of the backend implementation
            you are using.

        Returns
        -------
        ~pathlib.Path
            Path to the exported glb file.
        """
        if plotting_object is not None:
            self.plot(plotting_object, **plotting_options)

        # Depending on whether a name is provided, the file will be saved with the name
        # provided or with a default name (temp_glb). If a name is provided, the file will
        # be saved in the current working directory. If a path is provided, the file will be
        # saved in the provided path. We will also make sure that the file has the .glb extension,
        # and if not, we will add it.
        if filename is None:
            glb_filepath = Path("temp_glb.glb")
        else:
            glb_filepath = filename if isinstance(filename, Path) else Path(filename)
            if glb_filepath.suffix != ".glb":
                glb_filepath = glb_filepath.with_suffix(".glb")

        # Temporary gltf file to export the scene
        gltf_filepath = glb_filepath.with_suffix(".gltf")

        # Hide the axes before exporting the gltf and then export it
        self.backend._pl._scene.hide_axes()
        self.backend._pl._scene.export_gltf(gltf_filepath)

        # Convert the gltf to glb and remove the gltf file
        gltf2glb(gltf_filepath)
        Path.unlink(gltf_filepath)

        # Check if the file was exported correctly
        if not glb_filepath.exists():  # pragma: no cover
            raise FileNotFoundError(f"GLB file not found at {glb_filepath}. Export failed.")

        # Finally, return the path to the exported glb file
        return glb_filepath



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for creating a named selection."""

from typing import TYPE_CHECKING

from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.connection.conversions import grpc_point_to_point3d
from ansys.geometry.core.designer.beam import Beam
from ansys.geometry.core.designer.body import Body
from ansys.geometry.core.designer.designpoint import DesignPoint
from ansys.geometry.core.designer.edge import Edge
from ansys.geometry.core.designer.face import Face
from ansys.geometry.core.misc.auxiliary import (
    get_beams_from_ids,
    get_bodies_from_ids,
    get_edges_from_ids,
    get_faces_from_ids,
)

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.design import Design


class NamedSelection:
    """Represents a single named selection within the design assembly.

    This class synchronizes to a design within a supporting Geometry service instance.

    A named selection organizes one or more design entities together for common actions
    against the entire group.

    Parameters
    ----------
    name : str
        User-defined name for the named selection.
    design : Design
        Design instance to which the named selection belongs.
    grpc_client : GrpcClient
        Active supporting Geometry service instance for design modeling.
    bodies : list[Body], default: None
        All bodies to include in the named selection.
    faces : list[Face], default: None
        All faces to include in the named selection.
    edges : list[Edge], default: None
        All edges to include in the named selection.
    beams : list[Beam], default: None
        All beams to include in the named selection.
    design_points : list[DesignPoints], default: None
        All design points to include in the named selection.
    """

    def __init__(
        self,
        name: str,
        design: "Design",
        grpc_client: GrpcClient,
        bodies: list[Body] | None = None,
        faces: list[Face] | None = None,
        edges: list[Edge] | None = None,
        beams: list[Beam] | None = None,
        design_points: list[DesignPoint] | None = None,
        preexisting_id: str | None = None,
    ):
        """Initialize the ``NamedSelection`` class."""
        self._name = name
        self._design = design
        self._grpc_client = grpc_client

        # Create empty arrays if there are none of a type
        if bodies is None:
            bodies = []
        if faces is None:
            faces = []
        if edges is None:
            edges = []
        if beams is None:
            beams = []
        if design_points is None:
            design_points = []

        # Instantiate
        self._bodies = bodies
        self._faces = faces
        self._edges = edges
        self._beams = beams
        self._design_points = design_points

        # Store ids for later use... when verifying if the NS changed.
        self._ids_cached = {
            "bodies": [body.id for body in bodies],
            "faces": [face.id for face in faces],
            "edges": [edge.id for edge in edges],
            "beams": [beam.id for beam in beams],
            "design_points": [dp.id for dp in design_points],
        }

        if preexisting_id:
            self._id = preexisting_id
            return

        # All ids should be unique - no duplicated values
        ids = set()

        # Loop over all entities to get their ids
        for value in self._ids_cached.values():
            for entity_id in value:
                ids.add(entity_id)

        response = self._grpc_client.services.named_selection.create_named_selection(
            name=name, members=ids
        )
        self._id = response.get("id")

    @property
    def id(self) -> str:
        """ID of the named selection."""
        return self._id

    @property
    def name(self) -> str:
        """Name of the named selection."""
        return self._name

    @property
    def bodies(self) -> list[Body]:
        """All bodies in the named selection."""
        self.__verify_ns()
        if self._bodies is None:
            # Get all bodies from the named selection
            self._bodies = get_bodies_from_ids(self._design, self._ids_cached["bodies"])

        return self._bodies

    @property
    def faces(self) -> list[Face]:
        """All faces in the named selection."""
        self.__verify_ns()
        if self._faces is None:
            # Get all faces from the named selection
            self._faces = get_faces_from_ids(self._design, self._ids_cached["faces"])

        return self._faces

    @property
    def edges(self) -> list[Edge]:
        """All edges in the named selection."""
        self.__verify_ns()
        if self._edges is None:
            # Get all edges from the named selection
            self._edges = get_edges_from_ids(self._design, self._ids_cached["edges"])

        return self._edges

    @property
    def beams(self) -> list[Beam]:
        """All beams in the named selection."""
        self.__verify_ns()
        if self._beams is None:
            # Get all beams from the named selection
            self._beams = get_beams_from_ids(self._design, self._ids_cached["beams"])

        return self._beams

    @property
    def design_points(self) -> list[DesignPoint]:
        """All design points in the named selection."""
        self.__verify_ns()
        if self._design_points is None:
            # Get all design points from the named selection
            self._design_points = [
                DesignPoint(dp_id, f"dp: {dp_id}", grpc_point_to_point3d(dp_point))
                for dp_id, dp_point in self._ids_cached["design_points"]
            ]

        return self._design_points

    def __verify_ns(self) -> None:
        """Verify that the contents of the named selection are up to date."""
        if self._grpc_client.backend_version < (25, 2, 0):
            self._grpc_client.log.warning(
                "Accessing members of named selections is only"
                " consistent starting in version 2025 R2."
            )
            return

        # Get all entities from the named selection
        response = self._grpc_client.services.named_selection.get_named_selection(id=self._id)

        # Check if the named selection has changed
        ids = {
            "bodies": response.get("bodies"),
            "faces": response.get("faces"),
            "edges": response.get("edges"),
            "beams": response.get("beams"),
            "design_points": response.get("design_points"),
        }

        for key in ids:
            if ids[key] != self._ids_cached[key]:
                # Clear the cache for that specific entity
                setattr(self, f"_{key}", None)
                # Update the cache
                self._ids_cached[key] = ids[key]

    def __repr__(self) -> str:
        """Represent the ``NamedSelection`` as a string."""
        lines = [f"ansys.geometry.core.designer.selection.NamedSelection {hex(id(self))}"]
        lines.append(f"  Name                 : {self._name}")
        lines.append(f"  Id                   : {self._id}")
        lines.append(f"  N Bodies             : {len(self.bodies)}")
        lines.append(f"  N Faces              : {len(self.faces)}")
        lines.append(f"  N Edges              : {len(self.edges)}")
        lines.append(f"  N Beams              : {len(self.beams)}")
        lines.append(f"  N Design Points      : {len(self.design_points)}")
        return "\n".join(lines)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing methods and classes to detect the gRPC version used."""

from enum import Enum, unique

from google.protobuf.empty_pb2 import Empty
import grpc

# ATTEMPT v0 IMPORT
try:
    import ansys.api.dbu.v0.admin_pb2_grpc as dbu_v0_admin_pb2_grpc
except ImportError:
    dbu_v0_admin_pb2_grpc = None

# ATTEMPT v1 IMPORT
try:
    import ansys.api.dbu.v1.admin_pb2_grpc as dbu_v1_admin_pb2_grpc
except ImportError:
    dbu_v1_admin_pb2_grpc = None


@unique
class GeometryApiProtos(Enum):
    """Enumeration of the supported versions of the gRPC API protocol."""

    V0 = 0, dbu_v0_admin_pb2_grpc
    V1 = 1, dbu_v1_admin_pb2_grpc

    @staticmethod
    def get_latest_version() -> "GeometryApiProtos":
        """Get the latest version of the gRPC API protocol."""
        val = max(version.value[0] for version in GeometryApiProtos)
        return GeometryApiProtos.from_int_value(val)

    @staticmethod
    def from_int_value(version_int: int) -> "GeometryApiProtos":
        """Get the enumeration value from an integer."""
        for version in GeometryApiProtos:
            if version.value[0] == version_int:
                return version
        raise ValueError(f"Invalid version integer: {version_int}")

    @staticmethod
    def from_string(version_string: str) -> "GeometryApiProtos":
        """Get the enumeration value from a string."""
        for version in GeometryApiProtos:
            if version.name == version_string.capitalize():
                return version
        raise ValueError(f"Invalid version string: {version_string}")

    def verify_supported(self, channel: grpc.Channel) -> bool:
        """Check if the version is supported.

        Parameters
        ----------
        channel : grpc.Channel
            The gRPC channel to the server.

        Returns
        -------
        bool
            True if the server supports the version, otherwise False.

        Notes
        -----
        This method checks if the server supports the gRPC API protocol version.
        """
        pb2_grpc = self.value[1]
        if pb2_grpc is None:
            return False

        try:
            admin_stub = pb2_grpc.AdminStub(channel)
            admin_stub.Health(Empty())
            return True
        except grpc.RpcError:
            return False


def set_proto_version(
    channel: grpc.Channel, version: GeometryApiProtos | str | None = None
) -> "GeometryApiProtos":
    """Set the version of the gRPC API protocol used by the server.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    version : GeometryApiProtos | str | None
        The version of the gRPC API protocol to use. If None, the latest
        version is used.

    Returns
    -------
    GeometryApiProtos
        The version of the gRPC API protocol used by the server.
    """
    # Sanity check the input
    if isinstance(version, str):
        version = GeometryApiProtos.from_string(version)

    # Check the server supports the requested version (if specified)
    if version and not version.verify_supported(channel):
        raise ValueError(f"Server does not support the requested version: {version.name}")

    # If no version specified... Attempt to use all of them, starting
    # with the latest version
    if version is None:
        version = GeometryApiProtos.get_latest_version()
        while not version.verify_supported(channel):
            version = GeometryApiProtos.from_int_value(version.value[0] - 1)

    # Return the version
    return version



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import grpc

from .._version import GeometryApiProtos, set_proto_version
from .base.admin import GRPCAdminService
from .base.bodies import GRPCBodyService
from .base.coordinate_systems import GRPCCoordinateSystemService
from .base.dbuapplication import GRPCDbuApplicationService
from .base.designs import GRPCDesignsService
from .base.driving_dimensions import GRPCDrivingDimensionsService
from .base.edges import GRPCEdgesService
from .base.faces import GRPCFacesService
from .base.materials import GRPCMaterialsService
from .base.measurement_tools import GRPCMeasurementToolsService
from .base.named_selection import GRPCNamedSelectionService
from .base.parts import GRPCPartsService
from .base.prepare_tools import GRPCPrepareToolsService
from .base.repair_tools import GRPCRepairToolsService


class _GRPCServices:
    """
    Placeholder for the gRPC services (i.e. stubs).

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    version : GeometryApiProtos | str | None
        The version of the gRPC API protocol to use. If None, the latest
        version is used.

    Notes
    -----
    This class provides a unified interface to access the different
    gRPC services available in the Geometry API. It allows for easy
    switching between different versions of the API by using the
    `version` parameter in the constructor. The services are lazy-loaded
    to avoid unnecessary imports and to improve performance.
    """

    def __init__(self, channel: grpc.Channel, version: GeometryApiProtos | str | None = None):
        """
        Initialize the GRPCServices class.

        Parameters
        ----------
        channel : grpc.Channel
            The gRPC channel to the server.
        version : GeometryApiProtos | str | None
            The version of the gRPC API protocol to use. If None, the latest
            version is used.
        """
        # Set the proto version to be used
        self.version = set_proto_version(channel, version)
        self.channel = channel

        # Lazy load all the services
        self._admin = None
        self._bodies = None
        self._coordinate_systems = None
        self._dbu_application = None
        self._designs = None
        self._driving_dimensions = None
        self._edges = None
        self._faces = None
        self._materials = None
        self._measurement_tools = None
        self._named_selection = None
        self._parts = None
        self._prepare_tools = None
        self._repair_tools = None

    @property
    def admin(self) -> GRPCAdminService:
        """
        Get the admin service for the specified version.

        Returns
        -------
        GRPCAdminService
            The admin service for the specified version.
        """
        if not self._admin:
            # Import the appropriate admin service based on the version
            from .v0.admin import GRPCAdminServiceV0
            from .v1.admin import GRPCAdminServiceV1

            if self.version == GeometryApiProtos.V0:
                self._admin = GRPCAdminServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._admin = GRPCAdminServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._admin

    @property
    def bodies(self) -> GRPCBodyService:
        """
        Get the body service for the specified version.

        Returns
        -------
        GRPCBodyService
            The body service for the specified version.
        """
        if not self._bodies:
            # Import the appropriate body service based on the version
            from .v0.bodies import GRPCBodyServiceV0
            from .v1.bodies import GRPCBodyServiceV1

            if self.version == GeometryApiProtos.V0:
                self._bodies = GRPCBodyServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._bodies = GRPCBodyServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._bodies

    @property
    def coordinate_systems(self) -> GRPCCoordinateSystemService:
        """
        Get the coordinate systems service for the specified version.

        Returns
        -------
        GRPCCoordinateSystemService
            The coordinate systems service for the specified version.
        """
        if not self._coordinate_systems:
            # Import the appropriate coordinate systems service based on the version
            from .v0.coordinate_systems import GRPCCoordinateSystemServiceV0
            from .v1.coordinate_systems import GRPCCoordinateSystemServiceV1

            if self.version == GeometryApiProtos.V0:
                self._coordinate_systems = GRPCCoordinateSystemServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._coordinate_systems = GRPCCoordinateSystemServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._coordinate_systems

    @property
    def designs(self) -> GRPCDesignsService:
        """
        Get the designs service for the specified version.

        Returns
        -------
        GRPCDesignsService
            The designs service for the specified version.
        """
        if not self._designs:
            # Import the appropriate designs service based on the version
            from .v0.designs import GRPCDesignsServiceV0
            from .v1.designs import GRPCDesignsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._designs = GRPCDesignsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._designs = GRPCDesignsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._designs

    @property
    def dbu_application(self) -> GRPCDbuApplicationService:
        """
        Get the DBU application service for the specified version.

        Returns
        -------
        GRPCDbuApplicationService
            The DBU application service for the specified version.
        """
        if not self._dbu_application:
            # Import the appropriate DBU application service based on the version
            from .v0.dbuapplication import GRPCDbuApplicationServiceV0
            from .v1.dbuapplication import GRPCDbuApplicationServiceV1

            if self.version == GeometryApiProtos.V0:
                self._dbu_application = GRPCDbuApplicationServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._dbu_application = GRPCDbuApplicationServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._dbu_application

    @property
    def driving_dimensions(self) -> GRPCDrivingDimensionsService:
        """
        Get the driving dimensions service for the specified version.

        Returns
        -------
        GRPCDrivingDimensionsService
            The driving dimensions service for the specified version.
        """
        if not self._driving_dimensions:
            # Import the appropriate driving dimensions service based on the version
            from .v0.driving_dimensions import GRPCDrivingDimensionsServiceV0
            from .v1.driving_dimensions import GRPCDrivingDimensionsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._driving_dimensions = GRPCDrivingDimensionsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._driving_dimensions = GRPCDrivingDimensionsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._driving_dimensions

    @property
    def edges(self) -> GRPCEdgesService:
        """
        Get the edges service for the specified version.

        Returns
        -------
        GRPCEdgesService
            The edges service for the specified version.
        """
        if not self._edges:
            # Import the appropriate edges service based on the version
            from .v0.edges import GRPCEdgesServiceV0
            from .v1.edges import GRPCEdgesServiceV1

            if self.version == GeometryApiProtos.V0:
                self._edges = GRPCEdgesServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._edges = GRPCEdgesServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._edges

    @property
    def faces(self) -> GRPCFacesService:
        """
        Get the faces service for the specified version.

        Returns
        -------
        GRPCEdgesService
            The faces service for the specified version.
        """
        if not self._faces:
            # Import the appropriate faces service based on the version
            from .v0.faces import GRPCFacesServiceV0
            from .v1.faces import GRPCFacesServiceV1

            if self.version == GeometryApiProtos.V0:
                self._faces = GRPCFacesServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._faces = GRPCFacesServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._faces

    @property
    def materials(self) -> GRPCMaterialsService:
        """
        Get the materials service for the specified version.

        Returns
        -------
        GRPCMaterialsService
            The materials service for the specified version.
        """
        if not self._materials:
            # Import the appropriate materials service based on the version
            from .v0.materials import GRPCMaterialsServiceV0
            from .v1.materials import GRPCMaterialsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._materials = GRPCMaterialsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._materials = GRPCMaterialsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._materials

    @property
    def measurement_tools(self) -> GRPCMeasurementToolsService:
        """
        Get the measurement tools service for the specified version.

        Returns
        -------
        GRPCMeasurementToolsService
            The measurement tools service for the specified version.
        """
        if not self._measurement_tools:
            # Import the appropriate measurement tools service based on the version
            from .v0.measurement_tools import GRPCMeasurementToolsServiceV0
            from .v1.measurement_tools import GRPCMeasurementToolsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._measurement_tools = GRPCMeasurementToolsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._measurement_tools = GRPCMeasurementToolsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._measurement_tools

    @property
    def named_selection(self) -> GRPCNamedSelectionService:
        """
        Get the named selection service for the specified version.

        Returns
        -------
        GRPCNamedSelectionService
            The named selection service for the specified version.
        """
        if not self._named_selection:
            # Import the appropriate named selection service based on the version
            from .v0.named_selection import GRPCNamedSelectionServiceV0
            from .v1.named_selection import GRPCNamedSelectionServiceV1

            if self.version == GeometryApiProtos.V0:
                self._named_selection = GRPCNamedSelectionServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._named_selection = GRPCNamedSelectionServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._named_selection

    @property
    def parts(self) -> GRPCPartsService:
        """
        Get the parts service for the specified version.

        Returns
        -------
        GRPCPartsService
            The parts service for the specified version.
        """
        if not self._parts:
            # Import the appropriate parts service based on the version
            from .v0.parts import GRPCPartsServiceV0
            from .v1.parts import GRPCPartsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._parts = GRPCPartsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._parts = GRPCPartsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._parts

    @property
    def prepare_tools(self) -> GRPCPrepareToolsService:
        """
        Get the prepare tools service for the specified version.

        Returns
        -------
        GRPCPrepareToolsService
            The prepare tools service for the specified version.
        """
        if not self._prepare_tools:
            # Import the appropriate prepare tools service based on the version
            from .v0.prepare_tools import GRPCPrepareToolsServiceV0
            from .v1.prepare_tools import GRPCPrepareToolsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._prepare_tools = GRPCPrepareToolsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._prepare_tools = GRPCPrepareToolsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")

        return self._prepare_tools

    @property
    def repair_tools(self) -> GRPCRepairToolsService:
        """
        Get the repair tools service for the specified version.

        Returns
        -------
        RepairToolsServiceBase
            The repair tools service for the specified version.
        """
        if not self._repair_tools:
            from .v0.repair_tools import GRPCRepairToolsServiceV0
            from .v1.repair_tools import GRPCRepairToolsServiceV1

            if self.version == GeometryApiProtos.V0:
                self._repair_tools = GRPCRepairToolsServiceV0(self.channel)
            elif self.version == GeometryApiProtos.V1:  # pragma: no cover
                # V1 is not implemented yet
                self._repair_tools = GRPCRepairToolsServiceV1(self.channel)
            else:  # pragma: no cover
                # This should never happen as the version is set in the constructor
                raise ValueError(f"Unsupported version: {self.version}")
        return self._repair_tools



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the admin service implementation for v0."""

import warnings

import grpc
import semver

from ansys.geometry.core.errors import protect_grpc

from ..base.admin import GRPCAdminService
from .conversions import from_grpc_backend_type_to_backend_type


class GRPCAdminServiceV0(GRPCAdminService):
    """Admin service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    admin service. It is specifically designed for the v0 version of the
    Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.dbu.v0.admin_pb2_grpc import AdminStub

        self.stub = AdminStub(channel)

    @protect_grpc
    def get_backend(self, **kwargs) -> dict:  # noqa: D102
        # TODO: Remove this context and filter once the protobuf UserWarning is downgraded to INFO
        # https://github.com/grpc/grpc/issues/37609
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "Protobuf gencode version", UserWarning, "google.protobuf.runtime_version"
            )
            from google.protobuf.empty_pb2 import Empty

        # Create the request - assumes all inputs are valid and of the proper type
        request = Empty()

        # Call the gRPC service
        response = self.stub.GetBackend(request=request)

        # COMPATIBILITY HACK: retrieve the backend version -- for versions after 24R1
        if hasattr(response, "version"):
            ver = response.version
            backend_version = semver.Version(ver.major_release, ver.minor_release, ver.service_pack)
        else:  # pragma: no cover
            # If the version is not available, set a default version
            backend_version = semver.Version(24, 1, 0)

        # Convert the response to a dictionary
        return {
            "backend": from_grpc_backend_type_to_backend_type(response.type),
            "version": backend_version,
        }

    @protect_grpc
    def get_logs(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.dbu.v0.admin_pb2 import LogsRequest, LogsTarget, PeriodType

        # Create the request - assumes all inputs are valid and of the proper type
        request = LogsRequest(
            target=LogsTarget.CLIENT,
            period_type=PeriodType.CURRENT if not kwargs["all_logs"] else PeriodType.ALL,
            null_path=None,
            null_period=None,
        )

        # Call the gRPC service
        logs_generator = self.stub.GetLogs(request)
        logs: dict[str, str] = {}

        # Convert the response to a dictionary
        for chunk in logs_generator:
            if chunk.log_name not in logs:
                logs[chunk.log_name] = ""
            logs[chunk.log_name] += chunk.log_chunk.decode()

        return {"logs": logs}

    @protect_grpc
    def get_service_status(self, **kwargs) -> dict:  # noqa: D102
        from google.protobuf.empty_pb2 import Empty

        # Create the request - assumes all inputs are valid and of the proper type
        request = Empty()

        # Call the gRPC service
        response = self.stub.Health(request=request)

        # Convert the response to a dictionary
        return {"healthy": True if response.message == "I am healthy!" else False}
